__all__ = ()

from ...channel import ChannelType
from ...field_parsers import (
    bool_parser_factory, field_parser_factory, int_parser_factory, nullable_object_array_parser_factory,
    preinstanced_array_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, int_optional_putter_factory, nullable_entity_array_optional_putter_factory,
    nullable_field_optional_putter_factory, preinstanced_array_putter_factory,
)
from ...field_validators import (
    bool_validator_factory, nullable_object_array_validator_factory, preinstanced_array_validator_factory
)

from ..application_command_option_choice import ApplicationCommandOptionChoice

from .constants import (
    APPLICATION_COMMAND_OPTION_CHOICES_MAX, APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT,
    APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX, APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN,
    APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT, APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX,
    APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN, APPLICATION_COMMAND_OPTION_OPTIONS_MAX
)

# autocomplete

parse_autocomplete = bool_parser_factory('autocomplete', False)
put_autocomplete_into = bool_optional_putter_factory('autocomplete', False)
validate_autocomplete = bool_validator_factory('autocomplete', False)


# channel_types

parse_channel_types = preinstanced_array_parser_factory('channel_types', ChannelType)
put_channel_types_into = preinstanced_array_putter_factory('channel_types')
validate_channel_types = preinstanced_array_validator_factory('channel_types', ChannelType)


# choices

parse_choices = nullable_object_array_parser_factory('choices', ApplicationCommandOptionChoice)
put_choices_into = nullable_entity_array_optional_putter_factory('choices', ApplicationCommandOptionChoice)
validate_choices = nullable_object_array_validator_factory('choices', ApplicationCommandOptionChoice)


def validate_choices_postprocessed(choices, expected_choice_type):
    """
    Validates the given application command option `choices` field. 
    
    Parameters
    ----------
    choices : `None`, `iterable` of ``ApplicationCommandOptionChoice``
        The choices of the command for string or integer types. It's length can be in range [0:25].
    
    expected_choice_type : `type`
        The expected option value type.
    
    Returns
    -------
    choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
    
    Returns
    -------
    TypeError
        - If `choices`'s type is incorrect.
    ValueError
        - If `choices`'s value is incorrect.
    """
    choices = validate_choices(choices)
    
    if (choices is not None):
        for choice in choices:
            if not isinstance(choice.value, expected_choice_type):
                raise TypeError(
                    f'`choices` contains an element with a non-`{expected_choice_type.__name__}` value. '
                    f'Got {choice.__class__.__name__}; {choice!r}; choices = {choices!r}.'
                )
        
        choices = choices[:APPLICATION_COMMAND_OPTION_CHOICES_MAX]
    
    return choices


# default

parse_default = bool_parser_factory('default', False)
put_default_into = bool_optional_putter_factory('default', False)
validate_default = bool_validator_factory('default', False)


# max_length

def parse_max_length(data):
    """
    Parses out the `max_length` field from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Channel data.
    
    Returns
    -------
    max_length : `int`
    """
    max_length = data.get('max_length', None)
    if (max_length is None) or (max_length == APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX):
        max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT

    return max_length


def put_max_length_into(max_length, data, defaults):
    """
    Puts the `max_length`'s value into the given `data` json serializable object.
    
    Parameters
    ----------
    max_length : `bool`
        The `max_length` field value.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (max_length == 0):
        max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX
    
    data['max_length'] = max_length
    return data


def validate_max_length(max_length):
    """
    Validates an application command option's `max_length` field.
    
    Parameters
    ----------
    max_length : `None`, `int`
        The maximum input length allowed for this option.
        
        Only applicable for string options.
    
    Returns
    -------
    max_length : `int`
    
    Raises
    ------
    TypeError
        - If `max_length`'s type is incorrect.
    ValueError
        - If `max_length`'s value is incorrect.
    """
    if (max_length is None):
        return APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT
    
    if not isinstance(max_length, int):
        raise TypeError(
            f'`max_length` can be `None`, `int`, got {max_length.__class__.__name__}; {max_length!r}.'
        )
        
    if max_length == APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT:
        return APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT
    
    if max_length < APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN:
        max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN
    
    elif max_length >= APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX:
        max_length = APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT

    return max_length


# max_value

parse_max_value = field_parser_factory('max_value')
put_max_value_into = nullable_field_optional_putter_factory('max_value')

def validate_max_value(max_value):
    """
    Validates an application command option's `max_value` field.
    
    Parameters
    ----------
    max_value : `None`, `int`, `float`
        The maximum input value allowed for this option.
    
    Returns
    -------
    max_value : `None`, `int`, `float`
    
    Raises
    ------
    TypeError
        - If `max_value`'s type is incorrect.
    ValueError
        - If `max_value`'s value is incorrect.
    """
    if (max_value is None) or isinstance(max_value, (int, float)):
        return max_value
    
    raise TypeError(
        f'`max_value` can be either `None`, `int`, `float` depending on the option\'s type. '
        f'Got {max_value.__class__.__name__}; max_value = {max_value!r}.'
    )
    

def validate_max_value_postprocessed(max_value, expected_value_type):
    """
    Validates an application command option's `max_value` field.
    
    Parameters
    ----------
    max_value : `None`, `int`, `float`
        The maximum input value allowed for this option.
    expected_value_type : `type`
        The expected value type of `max_value`
    
    Returns
    -------
    max_value : `None`, `instance<expected_value_type>`
    
    Raises
    ------
    TypeError
        - If `max_value`'s type is incorrect.
    ValueError
        - If `max_value`'s value is incorrect.
        - if `max_value`'s type is not expected.
    """
    max_value = validate_max_value(max_value)
    if (max_value is None) or isinstance(max_value, expected_value_type):
        return max_value
    
    raise ValueError(
        f'`max_value` was expected either as `None`, `{expected_value_type.__name__}`, got '
        f'{max_value.__class__.__name__}; max_value = {max_value!r}.'
    )


# min_length

parse_min_length = int_parser_factory('min_length', APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT)
put_min_length_into = int_optional_putter_factory('min_length', APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN)


def validate_min_length(min_length):
    """
    Validates an application command option's `min_length` field.
    
    Parameters
    ----------
    min_length : `None`, `int`
        The minimum input length allowed for this option.
    
    Returns
    -------
    min_length : `int`
    
    Raises
    ------
    TypeError
        - If `min_length`'s type is incorrect.
    ValueError
        - If `min_length`'s value is incorrect.
        - If `min_length` is not applicable for the given type.
    """
    if (min_length is None):
        return APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT
    
    if not isinstance(min_length, int):
        raise TypeError(
            f'`min_length` can be `None`, `int`, got {min_length.__class__.__name__}; {min_length!r}.'
        )
    
    if min_length < APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN:
        min_length = APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN
    
    elif min_length > APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX:
        min_length = APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX

    return min_length


# min_value

parse_min_value = field_parser_factory('min_value')
put_min_value_into = nullable_field_optional_putter_factory('min_value')


def validate_min_value(min_value):
    """
    Validates an application command option's `min_value` field.
    
    Parameters
    ----------
    min_value : `None`, `int`, `float`
        The maximum input value allowed for this option.
        
    Returns
    -------
    min_value : `None`, `int`, `float`
    
    Raises
    ------
    TypeError
        - If `min_value`'s type is incorrect.
    ValueError
        - If `min_value`'s value is incorrect.
        - If `min_value` is not applicable for the given type.
    """
    if min_value is None or isinstance(min_value, (int, float)):
        return min_value
        
    raise TypeError(
        f'`min_value` can be either `None`, `int`, `float` depending on the option\'s type. '
        f'Got {min_value.__class__.__name__}; min_value = {min_value!r}.'
    )
    

def validate_min_value_postprocessed(min_value, expected_value_type):
    """
    Validates an application command option's `min_value` field.
    
    Parameters
    ----------
    min_value : `None`, `int`, `float`
        The minimal input value allowed for this option.
    expected_value_type : `type`
        The expected value type of `min_value`
    
    Returns
    -------
    min_value : `None`, `instance<expected_value_type>`
    
    Raises
    ------
    TypeError
        - If `min_value`'s type is incorrect.
    ValueError
        - If `min_value`'s value is incorrect.
        - if `min_value`'s type is not expected.
    """
    min_value = validate_min_value(min_value)
    if (min_value is None) or isinstance(min_value, expected_value_type):
        return min_value
    
    raise ValueError(
        f'`min_value` was expected either as `None`, `{expected_value_type.__name__}`, got '
        f'{min_value.__class__.__name__}; min_value = {min_value!r}.'
    )

# options

parse_options = nullable_object_array_parser_factory('options', NotImplemented, include = 'ApplicationCommandOption')
put_options_into = nullable_entity_array_optional_putter_factory(
    'options', NotImplemented, can_include_internals = False, include = 'ApplicationCommandOption'
)

validate_options = nullable_object_array_validator_factory(
    'options', NotImplemented, include = 'ApplicationCommandOption'
)


def validate_options_postprocessed(options):
    """
    Validates the given application command option `options` field.
    
    Parameters
    ----------
    options : `None`, `iterable` of ``ApplicationCommandOption``
        The parameters or sub-commands of the command option. It's length can be in range [0:25].
    
    Returns
    -------
    options : `None`, `tuple` of ``ApplicationCommandOption``
    
    Returns
    -------
    TypeError
        - If `options`'s type is incorrect.
    ValueError
        - If `options`'s value is incorrect.
        - If `options` is not applicable for the given type.
    """
    options = validate_options(options)
    if (options is not None):
        options = options[:APPLICATION_COMMAND_OPTION_OPTIONS_MAX]
    
    return options

# required

parse_required = bool_parser_factory('required', False)
put_required_into = bool_optional_putter_factory('required', False)
validate_required = bool_validator_factory('required', False)



__all__ = ('IntegrationMetadataDiscord',)

from scarletio import copy_docs

from ..integration_account import IntegrationAccount

from .base import IntegrationMetadataBase
from .fields import (
    parse_account__discord, parse_application, put_account_into__discord, put_application_into,
    validate_account__discord, validate_application
)


class IntegrationMetadataDiscord(IntegrationMetadataBase):
    """
    Integration metadata for discord integrations.
    
    Attributes
    ----------
    account : ``IntegrationAccount``, ``ClientUserBase``
        The invited bot.
    application : `None`, ``IntegrationApplication``
        The bot's application.
    """
    __slots__ = ('account', 'application',)
    
    def __new__(cls, *, account = ..., application = ...):
        """
        Creates a new discord integration metadata instance with the given parameters.
        
        Parameters
        ----------
        account : ``IntegrationAccount``, ``ClientUserBase``, Optional (Keyword only)
            The invited bot.
        application : `None`, ``IntegrationApplication``, Optional (Keyword only)
            The bot's application.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # account
        if account is ...:
            account = IntegrationAccount._create_empty()
        else:
            account = validate_account__discord(account)
        
        # application
        if application is ...:
            application = None
        else:
            application = validate_application(application)
        
        # Construct
        self = object.__new__(cls)
        self.account = account
        self.application = application
        return self
    
    
    @classmethod
    @copy_docs(IntegrationMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            account = keyword_parameters.pop('account', ...),
            application = keyword_parameters.pop('application', ...),
        )
    
    
    @copy_docs(IntegrationMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.account != other.account:
            return False
        
        if self.application != other.application:
            return False
        
        return True
    
    
    @copy_docs(IntegrationMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # account
        hash_value ^= hash(self.account)
        
        # application
        application = self.application
        if (application is not None):
            hash_value ^= hash(application)
        
        return hash_value
    
    
    @classmethod
    @copy_docs(IntegrationMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.account = IntegrationAccount._create_empty()
        self.application = None
        return self
    
    
    @classmethod
    @copy_docs(IntegrationMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.account = parse_account__discord(data)
        self.application = parse_application(data)
        return self
    
    
    @copy_docs(IntegrationMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_account_into__discord(self.account, data, defaults)
        put_application_into(self.application, data, defaults)
        return data
    
    
    @copy_docs(IntegrationMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.account = self.account
        new.application = self.application
        return new
    
    
    def copy_with(self, *, account = ..., application = ...):
        """
        Copies the discord integration metadata with the given fields.
        
        Creates a new discord integration metadata instance with the given parameters.
        
        Parameters
        ----------
        account : ``IntegrationAccount``, ``ClientUserBase``, Optional (Keyword only)
            The invited bot.
        application : `None`, ``IntegrationApplication``, Optional (Keyword only)
            The bot's application.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # account
        if account is ...:
            account = self.account
        else:
            account = validate_account__discord(account)
        
        # application
        if application is ...:
            application = self.application
        else:
            application = validate_application(application)
        
        self = object.__new__(type(self))
        self.account = account
        self.application = application
        return self
    
    
    @copy_docs(IntegrationMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            account = keyword_parameters.pop('account', ...),
            application = keyword_parameters.pop('application', ...),
        )



__all__ = ()

from ...application_command import ApplicationCommand
from ...application_command.application_command.constants import (
    APPLICATION_COMMAND_NAME_LENGTH_MAX, APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from ...bases import DiscordEntity
from ...component import ComponentType
from ...component.shared_constants import CUSTOM_ID_LENGTH_MAX
from ...field_parsers import (
    entity_id_parser_factory, force_string_parser_factory, nullable_object_array_parser_factory,
    nullable_sorted_array_parser_factory, nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, force_string_putter_factory,
    nullable_object_array_optional_putter_factory, nullable_string_array_optional_putter_factory,
    preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_entity_validator_factory,
    nullable_object_array_validator_factory, nullable_string_array_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)

from ..interaction_component import InteractionComponent
from ..interaction_option import InteractionOption
from ..resolved import Resolved

# component_type

parse_component_type = preinstanced_parser_factory('component_type', ComponentType, ComponentType.none)
put_component_type_into = preinstanced_putter_factory('component_type')
validate_component_type = preinstanced_validator_factory('component_type', ComponentType)

# components

parse_components = nullable_object_array_parser_factory('components', InteractionComponent)
put_components_into = nullable_object_array_optional_putter_factory('components')
validate_components = nullable_object_array_validator_factory('components', InteractionComponent)

# custom_id

parse_custom_id = nullable_string_parser_factory('custom_id')
put_custom_id_into = url_optional_putter_factory('custom_id')
validate_custom_id = nullable_string_validator_factory('custom_id', 0, CUSTOM_ID_LENGTH_MAX)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id', ApplicationCommand)

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory(
    'name', APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX
)

# options

parse_options = nullable_object_array_parser_factory('options', InteractionOption)
put_options_into = nullable_object_array_optional_putter_factory('options')
validate_options = nullable_object_array_validator_factory('options', InteractionOption)

# resolved

def parse_resolved(data, interaction_event):
    """
    Parsers out a resolved object from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Interaction metadata data.
    
    interaction_event : ``InteractionEvent``
        Parent interaction event instance.
    
    Returns
    -------
    resolved : `None`, ``Resolved``
    """
    resolved_data = data.get('resolved', None)
    if (resolved_data is not None) and resolved_data:
        return Resolved.from_data(resolved_data, interaction_event)


def put_resolved_into(resolved, data, defaults, *, interaction_event = None):
    """
    Puts the given `resolved` into the given interaction metadata data.
    
    Parameters
    ----------
    resolved  : `None`, ``Resolved``
        
    data : `dict` of (`str`, `object`) items
        Interaction metadata data.
    
    defaults : `bool`
        Whether default field values should be included as well.
    
    interaction_event : ``InteractionEvent`` = `None`, Optional (Keyword only)
        The respective guild's identifier to use for handing user guild profiles.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    # Cpython devs: "We do not need goto in python".
    # Also them: *uSeS gOtO TwIcE iN EveRy C fUnCTiOn*.
    while True:
        if (resolved is None):
            if not defaults:
                break
            
            resolved_data = {}
        
        else:
            resolved_data = resolved.to_data(defaults = defaults, interaction_event = interaction_event)
        
        data['resolved'] = resolved_data
        break
    
    return data


validate_resolved = nullable_entity_validator_factory('resolved', Resolved)

# target_id

parse_target_id = entity_id_parser_factory('target_id')
put_target_id_into = entity_id_optional_putter_factory('target_id')
validate_target_id = entity_id_validator_factory('target_id', DiscordEntity)

# values

parse_values = nullable_sorted_array_parser_factory('values')
put_values_into = nullable_string_array_optional_putter_factory('values')
validate_values = nullable_string_array_validator_factory('values')



__all__ = ()

APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN = 1
APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX = 100
APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN = 0
APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MAX = 100



"""
Checks can be added to commands or to categories to limit their usage to set users or places.


Checks can be either used as a decorator:

```py
from hata.ext.commands import checks

@Momiji.commands
@checks.owner_only()
async def knock_knock():
    return 'Awu!'
```

Or as a `.commands` parameter:

```py
from hata.ext.commands import checks

@Momiji.commands(checks=checks.owner_only())
async def knock_knock():
    return 'Awu!'
```

The implemented checks are the following:

+--------------------------------+-----------------+--------------------------------------------------------------+
| Name                           | Extra parameter | Description                                                  |
+================================+=================+==============================================================+
| announcement_channel_only      | N/A             | Whether the message's channel is an announcement channel.    |
+--------------------------------+-----------------+--------------------------------------------------------------+
| booster_only                   | N/A             | Whether the user boosts the respective guild.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| bot_account_only               | N/A             | Whether the message's author is a bot account.               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| client_only                    | N/A             | Whether the message was sent by a ``Client``.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| custom                         | function        | Custom checks, to wrap a given `function`. (Can be async.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_only                     | N/A             | Whether the message was sent to a guild channel.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_owner_only               | N/A             | Whether the message's author is the guild's owner.           |
|                                |                 | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_any_role                   | *roles          | Whether the message's author has any of the given roles.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_client_guild_permissions   | permissions,    | Whether the client has the given permissions at the guild.   |
|                                | **kwargs        | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_client_permissions         | permissions,    | Whether the client has the given permissions at the channel. |
|                                | **kwargs        |                                                              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_guild_permissions          | permissions,    | Whether the message's author has the given permissions at    |
|                                | **kwargs        | the guild. (Fails in private channels.)                      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_permissions                | permissions,    | Whether the message's author has the given permissions at    |
|                                | **kwargs        | the channel.                                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_role                       | role            | Whether the message's author has the given role.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_category                | *categories     | Whether the message was sent into a channel, what's category |
|                                |                 | is any of the specified ones.                                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_channel                 | *channels       | Whether the message was sent to any of the given channels.   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_guild                   | *guilds         | Whether the message was sent to any of the given guilds.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_category                    | category        | Whether the message was sent into a channel, what's category |
|                                |                 | is the specified one.                                        |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_channel                     | channel         | Whether the message's channel is the given one.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_guild                       | guild           | Whether the message guild is the given one.                  |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_in_voice                    | N/A             | Whether the user is in a voice channel in the respective     |
|                                |                 | guild.                                                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| release_at                     | release_at,     | Whether the command is already released. Users with the      |
|                                | *roles          | given roles and the bot owners bypass the check.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| nsfw_channel_only              | N/A             | Whether the message's channel is nsfw.                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_only                     | N/A             | Whether the message's author is an owner of the client.      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_guild_owner_only      | N/A             | `owner_only`, `guild_owner` (Fails in private channels.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_any_role          | *roles          | `owner_only`, `has_any_role`                               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_guild_permissions | permissions,    | `owner_only`, `has_guild_permissions`                      |
|                                | **kwargs        | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_permissions       | permissions,    | `owner_only`, `has_permissions`                            |
|                                | **kwargs        |                                                              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_role              | role            | `owner_only`, `has_role`                                   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| private_only                   | N/A             | Whether the message's channel is a private channel.          |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_only              | N/A             | Whether the message's author is a user account.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_or_client_only    | N/A             | Whether the message's author is a user account or a          |
|                                |                 | ``Client``.                                                  |
+--------------------------------+-----------------+--------------------------------------------------------------+

To handle a failed check, do:

```py
from hata.ext.commands import checks, CommandCheckError

@Momiji.commands
@checks.owner_only()
async def knock_knock():
    return 'Awu!'

@knock_knock.error
async def error_handler(ctx, exception):
    # `exception` is ``CommandCheckError`` if a check fails.
    if isinstance(err, CommandCheckError):
        await ctx.send('Owner only!')
        # Return `True` if we handled the exception and further error handles should not be called.
        return True
    
    return False
```
"""

__all__ = (
    'CheckBase',
    'announcement_channel_only',
    'booster_only',
    'bot_account_only',
    'client_only',
    'custom',
    'guild_only',
    'guild_owner_only',
    'has_any_role',
    'has_client_guild_permissions',
    'has_client_permissions',
    'has_guild_permissions',
    'has_permissions',
    'has_role',
    'is_any_category',
    'is_any_channel',
    'is_any_guild',
    'is_category',
    'is_channel',
    'is_guild',
    'is_in_voice',
    'release_at',
    'nsfw_channel_only',
    'owner_only',
    'owner_or_guild_owner_only',
    'owner_or_has_any_role',
    'owner_or_has_guild_permissions',
    'owner_or_has_permissions',
    'owner_or_has_role',
    'private_only',
    'user_account_only',
    'user_account_or_client_only',
)

from datetime import datetime
from functools import partial as partial_func

from scarletio import CallableAnalyzer, Task, copy_docs, export

from ...discord.bases import instance_or_id_to_instance, instance_or_id_to_snowflake
from ...discord.channel import Channel
from ...discord.client import Client
from ...discord.core import KOKORO
from ...discord.guild import Guild
from ...discord.permission import Permission
from ...discord.role import Role
from ...discord.user import ClientUserBase
from ...discord.utils import datetime_to_id

from .wrappers import CommandCheckWrapper


def _convert_permission(permission):
    """
    Validates the given `permission`.
    
    Parameters
    ----------
    permission : `None`, ``Permission``, `int`.
        Permission to validate.
    
    Returns
    -------
    permissions : ``Permission``
    
    Raises
    ------
    TypeError
        If `permissions` was not given as `None`, ``Permission`` nor as `int`.
    """
    if permission is None:
        permission = Permission()
    else:
        if type(permission) is Permission:
            pass
        elif isinstance(permission, int):
            permission = Permission(permission)
        else:
            raise TypeError(
                f'`permission` can be `None`, `{Permission.__name__}`, `int`, got '
                f'{permission.__class__.__name__}; {permission!r}.'
            )
    
    return permission


class CheckMeta(type):
    """
    Check metaclass for collecting their `__slots__` in a `__all_slot__` class attribute.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type`
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``CheckMeta``
        """
        if class_parents:
            parent = class_parents[0]
            inherited_slots = getattr(parent, '__all_slot__', None)
        else:
            inherited_slots = None
        
        new_slots = class_attributes.get('__slots__', None)
        
        final_slots = []
        if (inherited_slots is not None):
            final_slots.extend(inherited_slots)
        
        if (new_slots is not None):
            final_slots.extend(new_slots)
        
        class_attributes['__all_slot__'] = tuple(final_slots)
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


@export
class CheckBase(metaclass=CheckMeta):
    """
    Base class for checks.
    """
    __slots__ = ()
    def __new__(cls):
        """
        Creates a new check instance.
        
        Subclasses should overwrite it.
        """
        return object.__new__(cls)
    
    async def __call__(client, context):
        """
        Returns whether the check's condition passes.
        
        Subclasses should overwrite it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        context : ``CommandContext``
            Processing context for the respective command.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
    
    def __repr__(self):
        """Returns the check's representation."""
        repr_parts = [
            self.__class__.__name__,
            '(',
        ]
        
        slots = self.__all_slot__
        limit = len(slots)
        if limit:
            index = 0
            while True:
                name = slots[index]
                index += 1
                if name.startswith('_'):
                    if index == limit:
                        break
                    else:
                        continue
                
                # case of `channel_id`, `guild_id`
                if name.endswith('id'):
                    display_name = name[:-3]
                # case of `channel_ids`, `guild_ids`
                elif name.endswith('ids'):
                    display_name = name[:-4]
                else:
                    display_name = name
                
                repr_parts.append(display_name)
                repr_parts.append('=')
                attr = getattr(self,name)
                repr_parts.append(repr(attr))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)
    
    def __invert__(self):
        """Inverts the check's condition returning a new check."""
        return CheckInvert(self)
    
    def __or__(self, other):
        """Connects the two check with `or` relation."""
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if type(self) is CheckBase:
            return other
        
        return CheckOrRelation(self, other)
    
    def __and__(self, other):
        """Connects the two checks which `and` relation."""
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is CheckBase:
            return self
        
        return CheckAndRelation(self, other)


class CheckInvert(CheckBase):
    """
    Inverts the wrapped check's result.
    
    Attributes
    ----------
    check : ``CheckBase``
        The check to invert.
    """
    __slots__ = ('check', )
    
    def __new__(cls, check):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        check : ``CheckBase``
            The check to invert.

        Raises
        ------
        TypeError
            If `check` was not given as ``CheckBase``.
        """
        if not isinstance(check, CheckBase):
            raise TypeError(
                f'`check` can be `{CheckBase.__name__}`, got {check.__class__.__name__}; {check!r}.'
            )
        
        self = object.__new__(cls)
        self.check = check
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        return not await self.check(context)


class CheckRelationBase(CheckBase):
    """
    Base class for relation checks.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ('checks', )
    
    def __new__(cls, *checks):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *checks : ``CheckBase``
            The check to invert.

        Raises
        ------
        TypeError
            If a `check` was not given as ``CheckBase``.
        """
        for check in checks:
            if not isinstance(check, CheckBase):
                raise TypeError(
                    f'`check` can be `{CheckBase.__name__}`, got {check.__class__.__name__}; {check!r}.'
                )
        
        self = object.__new__(cls)
        self.checks = checks
        return self


class CheckOrRelation(CheckRelationBase):
    """
    Connects checks with `or` relation.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        for check in self.checks:
            if await check(context):
                return True
        
        return False


class CheckAndRelation(CheckRelationBase):
    """
    Connects checks with `and` relation.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        for check in self.checks:
            if not await check(context):
                return False
        
        return True



class CheckSingleBase(CheckBase):
    """
    Base class for single condition checks without attributes.
    """
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        return CheckAndRelation(self, other)


class CheckIsOwner(CheckSingleBase):
    """
    Checks whether the command was called by the client's owner.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.client.is_owner(context.message.author):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        # Let the other check decide
        return other | self
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is other_type:
            return self
        
        if issubclass(other, CheckIsOwner):
            if (type(self) is CheckIsOwner):
                return self
            
            if type(other) is CheckIsOwner:
                return other
        
        return CheckAndRelation(self, other)


class CheckHasRoleBase(CheckBase):
    """
    Base class for role checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasRoleBase):
            return CheckOrRelation(self, other)
        
        roles = {*self._iter_roles(), *other._iter_roles()}
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        if owner_allowed:
            check_type = CheckHasRoleOrIsOwner
        else:
            check_type = HasAnyRoleCheckOrRelationIsOwner
        
        return check_type(*roles)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckHasRoleBase):
            return CheckAndRelation(self, other)
        
        roles = {*self._iter_roles()}&{*other._iter_roles()}
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        if owner_required:
            check_type = CheckHasRoleOrIsOwner
        else:
            check_type = HasAnyRoleCheckOrRelationIsOwner
        
        return check_type(*roles)
    
    
    def _iter_roles(self):
        """
        Iterates the roles of the check.
        
        This method is a generator.
        
        Yields
        ------
        role : ``Role``
        """
        return
        yield


class CheckHasRole(CheckHasRoleBase):
    """
    Checks whether the message's author has the given role.
    
    Attributes
    ----------
    role : ``Role``
        The legend itself.
    """
    __slots__ = ('role', )
    
    def __new__(cls, role):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        role : `str`, `int`, ``Role``
            The role what the message's author should have.
        
        Raises
        ------
        TypeError
            If `role` was not given neither as ``Role``, `str`, `int`.
        ValueError
            If `role` was given as `str`, `int`, but not as a valid snowflake, so ``Role``
            cannot be precreated with it.
        """
        role = instance_or_id_to_instance(role, Role, 'role')
        
        self = object.__new__(cls)
        self.role = role
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.author.has_role(self.role):
            return True
        
        return False
    
    @copy_docs(CheckHasRoleBase._iter_roles)
    def _iter_roles(self):
        yield self.role

class CheckHasRoleOrIsOwner(CheckHasRole, CheckIsOwner):
    """
    Checks whether the message's author has the given role, or if it the client's owner.
    
    Attributes
    ----------
    role : ``Role``
        The legend itself.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if user.has_role(self.role):
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasAnyRole(CheckHasRoleBase):
    """
    Checks whether the message's author has any of the given roles.
    
    Attributes
    ----------
    roles : `set` of ``Role``
        The legends themselves.
    """
    __slots__ = ('roles', )
    
    def __new__(cls, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *roles : `str`, `int`, ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If a role was not given neither as ``Role``, `str`, `int`.
        ValueError
            If a role was given as `str`, `int`, but not as a valid snowflake, so a ``Role``
            cannot be precreated with it.
        """
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        roles_processed_length = len(roles_processed)
        if roles_processed_length == 0:
            return CheckBase()
        
        if roles_processed_length == 1:
            return CheckHasRole(roles_processed.pop())
        
        self = object.__new__(cls)
        self.roles = roles_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        return False
    
    @copy_docs(CheckHasRoleBase._iter_roles)
    def _iter_roles(self):
        yield from self.roles


class HasAnyRoleCheckOrRelationIsOwner(CheckHasAnyRole, CheckIsOwner):
    """
    Checks whether the message's author has any of the given roles, or whether is it the client's owner.
    
    Attributes
    ----------
    roles : `set` of ``Role``
        The roles from what the user should have at least 1.
    """
    __slots__ = ()
    
    def __new__(cls, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *roles : `str`, `int`, ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If an element of role was not given neither as ``Role``, `str`, `int`.
        ValueError
            If a role was given as `str`, `int`, but not as a valid snowflake, so a ``Role``
            cannot be precreated with it.
        """
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        roles_processed_length = len(roles_processed)
        if roles_processed_length == 0:
            return CheckIsOwner()
        
        if roles_processed_length == 1:
            return CheckHasRoleOrIsOwner(roles_processed.pop())
        
        self = object.__new__(cls)
        self.roles = roles_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckIsInGuild(CheckSingleBase):
    """
    Checks whether the command was called from a guild.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        return context.message.channel.is_in_group_guild()
    
    @copy_docs(CheckBase.__invert__)
    def __invert__(self):
        return CheckIsInPrivate()


class CheckIsInPrivate(CheckSingleBase):
    """
    Checks whether the command was used inside of a private channel.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        return context.message.channel.is_in_group_private()
    
    @copy_docs(CheckBase.__invert__)
    def __invert__(self):
        return CheckIsInGuild()


class CheckIsGuildOwner(CheckBase):
    """
    Checks whether the command was called by the local guild's owner.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.context
        guild = message.guild
        if guild is None:
            return False
        
        if guild.owner_id == message.author.id:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is not CheckIsOwner:
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner):
            return self
        
        return CheckIsGuildOwnerOrIsOwner()


class CheckIsGuildOwnerOrIsOwner(CheckIsGuildOwner, CheckIsOwner):
    """
    Checks whether a message was sent by the message's guild's owner or by the client's owner.
    
    > Guild check is always applied.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.context
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        if guild.owner_id == user.id:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasPermissionBase(CheckBase):
    """
    Base class for checking permissions.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    def __new__(cls, permission = None, **kwargs):
        """
        Creates a check, which will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : `None`, ``Permission``, `int` = `None`, Optional
            The permission, which the respective user should have. Defaults to `None`
        **kwargs : Keyword parameters
            `permission-name` - `bool` relations.
        
        Raises
        ------
        LookupError
            If a keyword is not a valid permission name.
        TypeError
            If `permission` was not given neither as `None`, ``Permission`` nor as `int`.
        """
        permission = _convert_permission(permission)
        permission = permission.update_by_keys(**kwargs)
        
        if not permission:
            if issubclass(cls, CheckIsOwner):
                permission_type = CheckIsOwner
            else:
                permission_type = CheckBase
            
            return permission_type()
        
        self = object.__new__(cls)
        self.permission = permission
        return self


class CheckHasPermission(CheckHasPermissionBase):
    """
    Checks whether the message's author has the given permissions in the message's channel.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        if message.channel.permissions_for(message.author) >= self.permission:
            return True
        
        return False

    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasPermission):
            return CheckAndRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        permission = Permission(self.permission & other.permission)
        
        if owner_allowed:
            check_type = CheckHasPermissionOrIsOwner
        else:
            check_type = CheckHasPermission
        
        return check_type(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasPermission):
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        permission = Permission(self.permission | other.permission)
        if permission:
            if owner_required:
                check_type = CheckHasPermissionOrIsOwner
            else:
                check_type = CheckHasPermission
            
            return check_type(permission)
        else:
            if owner_required:
                check_type = CheckIsOwner
            else:
                check_type = CheckBase
            
            return check_type()



class CheckHasPermissionOrIsOwner(CheckHasPermission, CheckIsOwner):
    """
    Checks whether the message's author is the client's owner or has the given permissions in the message's channel.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        user = message.author
        if message.channel.permissions_for(user) >= self.permission:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasGuildPermission(CheckHasPermissionBase):
    """
    Checks whether the message's author has the given permissions in the message's guild.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        if guild.permissions_for(message.author) >= self.permission:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasGuildPermission):
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        permission = Permission(self.permission & other.permission)
        
        if owner_allowed:
            check_type = CheckHasGuildPermissionOrIsOwner
        else:
            check_type = CheckHasGuildPermission
        
        return check_type(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasGuildPermission):
            return CheckAndRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        permission = Permission(self.permission | other.permission)
        if permission:
            if owner_required:
                check_type = CheckHasGuildPermissionOrIsOwner
            else:
                check_type = CheckHasGuildPermission
            
            return check_type(permission)
        else:
            if owner_required:
                check_type = CheckIsOwner
            else:
                check_type = CheckBase
            
            return check_type()


class CheckHasGuildPermissionOrIsOwner(CheckHasGuildPermission, CheckIsOwner):
    """
    Checks whether the message's author has the given permissions in the message's guild.
    
    > Guild check is always applied.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        if guild.permissions_for(user) >= self.permission:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasClientPermission(CheckHasPermissionBase):
    """
    Checks whether the client has the given permissions in the message's channel.
    
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.channel.cached_permissions_for(context.client) >= self.permission:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if type(self) is not other_type:
            return CheckOrRelation(self, other)
        
        permission = Permission(self.permission & other.permission)
        
        return type(self)(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is not other_type:
            return CheckOrRelation(self, other)
        
        permission = Permission(self.permission | other.permission)
        
        if not permission:
            return CheckBase()
        
        return type(self)(permission)


class CheckHasClientGuildPermission(CheckHasClientPermission):
    """
    Checks whether the client has the given permissions in the message's guild.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if guild.cached_permissions_for(context.client) >= self.permission:
            return True
        
        return False


class CheckIsGuildBase(CheckBase):
    """
    Base class fro guild checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsGuildBase):
            return CheckOrRelation(self, other)
        
        guild_ids = {*self._iter_guild_ids(), *other._iter_guild_ids()}
        
        return CheckIsAnyGuild(*guild_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsGuildBase):
            return CheckAndRelation(self, other)
    
        guild_ids = {*self._iter_guild_ids()}&{*other._iter_guild_ids()}
        
        return CheckIsAnyGuild(*guild_ids)
    
    def _iter_guild_ids(self):
        """
        Iterates the guild ids of the check.
        
        This method is a generator.
        
        Yields
        ------
        guild_id : ``int``
        """
        return
        yield
    
    
class CheckIsGuild(CheckIsGuildBase):
    """
    Checks whether the message was sent to the given guild.
    
    Attributes
    ----------
    guild_id : `int`
        The respective guild's id.
    """
    __slots__ = ('guild_id', )
    
    def __new__(cls, guild):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        guild : `str`, `int`, ``Guild``
            The guild where the message should be sent.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, `str`, `int`.
        ValueError
            If `guild` was given as `str`, `int`, but not as a valid snowflake.
        """
        guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if (guild.id == self.guild_id):
            return True
        
        return False
    
    @copy_docs(CheckIsGuildBase._iter_guild_ids)
    def _iter_guild_ids(self):
        yield self.guild_id


class CheckIsAnyGuild(CheckIsGuildBase):
    """
    Checks whether the message was sent into any of the given guilds.
    
    Attributes
    ----------
    guild_ids : `set` of `int`
        The respective guilds' identifiers.
    """
    __slots__ = ('guild_ids', )
    
    def __new__(cls, *guilds):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *guilds : `str`, `int`, ``Guild``
            The guilds where the message should be sent.
        
        Raises
        ------
        TypeError
            If a guild was not given neither as ``Guild``, `str`, `int`.
        ValueError
            If a guild was given as `str`, `int`, but not as a valid snowflake.
        """
        guild_ids_processed = set()
        
        for guild in guilds:
            guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
            guild_ids_processed.add(guild_id)
        
        guild_ids_processed_length = len(guild_ids_processed)
        if guild_ids_processed_length == 0:
            return CheckBase()
        
        if guild_ids_processed_length == 1:
            return CheckIsGuild(guild_ids_processed.pop())
        
        self = object.__new__(cls)
        self.guild_ids = guild_ids_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if (guild.id in self.guild_ids):
            return True
        
        return False
    
    @copy_docs(CheckIsGuildBase._iter_guild_ids)
    def _iter_guild_ids(self):
        yield from self.guild_ids


class CheckCustom(CheckBase):
    """
    Does a custom check.
    
    Attributes
    ----------
    _is_async : `bool`
        Whether ``.check`` is async.
    check : `callable`
        The custom check's function.
    """
    __slots__ = ('_is_async', 'check', )

    def __new__(cls, check):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        check : `callable`
            The custom check which should pass. Can be async.
        
        Raises
        ------
        TypeError
            - If `check` was not given as an `callable`.
            - If `check` accepts more or less non reserved positional non default parameters.
        
        Notes
        -----
        Only `int`-s are evaluated to boolean.
        """
        analyzer = CallableAnalyzer(check)
        non_reserved_positional_parameter_count = analyzer.get_non_reserved_positional_parameter_count()
        if  non_reserved_positional_parameter_count != 1:
            raise TypeError(
                f'`check` should accept `1` non reserved, positional, non default parameters, meanwhile it accepts '
                f'{non_reserved_positional_parameter_count}, got {check!r}.'
            )
        
        is_async = analyzer.is_async()
        
        self = object.__new__(cls)
        self.check = check
        self._is_async = is_async
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        try:
            result = self.check(context)
            if self._is_async:
                result = await result
        except GeneratorExit:
            raise
        
        except BaseException as err:
            client = context.client
            Task(KOKORO, client.events.error(client, repr(self), err))
            return False
        
        if result is None:
            return False
        
        if isinstance(result, int) and result:
            return True
        
        return False


class CheckIsChannelBase(CheckBase):
    """
    Base class for channel checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsChannelBase):
            return CheckOrRelation(self, other)
        
        channel_ids = {*self._iter_channel_ids(), *other._iter_channel_ids()}
        
        return CheckIsAnyChannel(*channel_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsChannelBase):
            return CheckAndRelation(self, other)
    
        channel_ids = {*self._iter_channel_ids()}&{*other._iter_channel_ids()}
        
        return CheckIsAnyChannel(*channel_ids)
    
    def _iter_channel_ids(self):
        """
        Iterates the channel id-s of the check.
        
        This method is a generator.
        
        Yields
        ------
        channel_id : ``int``
        """
        return
        yield


class CheckIsChannel(CheckIsChannelBase):
    """
    Checks whether the message was sent to the given channel.
    
    Attributes
    ----------
    channel_id : `int`
        The respective channel's id.
    """
    __slots__ = ('channel_id', )
    
    def __new__(cls, channel):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        channel : `str`, `int`, ``Guild``
            The guild where the message should be sent.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, `str`, `int`.
        ValueError
            If `channel` was given as `str`, `int`, but not as a valid snowflake.
        """
        channel_id = instance_or_id_to_snowflake(channel, Channel, 'channel')
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if (context.message.channel.id == self.channel_id):
            return True
        
        return False
    
    @copy_docs(CheckIsChannelBase._iter_channel_ids)
    def _iter_channel_ids(self):
        yield self.channel_id


class CheckIsAnyChannel(CheckIsChannelBase):
    """
    Checks whether the message was sent into any of the given channels.
    
    Attributes
    ----------
    channel_ids : `set` of `int`
        The respective channels' identifiers.
    """
    __slots__ = ('channel_ids', )
    
    def __new__(cls, *channels):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *channels : `str`, `int`, ``Channel``
            The channel where the message should be sent.
        
        Raises
        ------
        TypeError
            If a channel was not given neither as ``Channel``, `str`, `int`.
        ValueError
            If a channel was given as `str`, `int`, but not as a valid snowflake.
        """
        channel_ids_processed = set()
        
        for channel in channels:
            channel_id = instance_or_id_to_snowflake(channel, Channel, 'guild')
            channel_ids_processed.add(channel_id)
        
        channel_ids_processed_length = len(channel_ids_processed)
        if channel_ids_processed_length == 0:
            return CheckBase()
        
        if channel_ids_processed_length == 1:
            return CheckIsChannel(channel_ids_processed.pop())
        
        self = object.__new__(cls)
        self.channel_ids = channel_ids_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if (context.message.channel.id in self.channel_ids):
            return True
        
        return False
    
    @copy_docs(CheckIsChannelBase._iter_channel_ids)
    def _iter_channel_ids(self):
        yield from self.channel_ids


class CheckIsNsfwChannel(CheckSingleBase):
    """
    Checks whether the message was sent to an nsfw channel.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        return context.message.channel.nsfw


class CheckIsAnnouncementChannel(CheckSingleBase):
    """
    Checks whether the message was sent to an announcement channel.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.channel.type == 5:
            return True
        
        return False


class CheckIsInVoice(CheckSingleBase):
    """
    Checks whether the message's author is in a voice channel in the respective guild.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        if message.author.id in guild.voice_states:
            return True
        
        return False


class CheckIsBooster(CheckSingleBase):
    """
    Checks whether the message's author boosts the respective guild.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return
        
        try:
            profile = message.author.guild_profiles[guild.id]
        except KeyError:
            return False
        
        if profile.boosts_since is None:
            return False
        
        return True


class CheckIsClient(CheckSingleBase):
    """
    Check whether the message was sent by a client.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if isinstance(context.message.author, Client):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is CheckUserAccount:
            return CheckIsUserAccountOrIsClient()
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsUserAccountOrIsClient):
            return self
        
        return CheckAndRelation(self, other)


class CheckUserAccount(CheckSingleBase):
    """
    Checks whether the message was sent by an user account.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if not context.message.author.bot:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is CheckIsClient:
            return CheckIsUserAccountOrIsClient()
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsUserAccountOrIsClient):
            return self
        
        return CheckAndRelation(self, other)


class CheckBotAccount(CheckSingleBase):
    """
    Checks whether the message was sent by a bot account.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if user.bot and isinstance(user, ClientUserBase):
            return True
        
        return False


class CheckIsUserAccountOrIsClient(CheckIsClient, CheckUserAccount, CheckSingleBase):
    """
    Checks whether the message was sent by a user account or by a client.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if not user.bot:
            return True
        
        if isinstance(user, Client):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if (other_type is type(self)) or (other_type is CheckIsClient) or (other_type is CheckUserAccount):
            return self
        
        return CheckOrRelation(self, other)

    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsClient) or (other_type is CheckUserAccount):
            return other
        
        return CheckAndRelation(self, other)


class CheckIsCategoryBase(CheckBase):
    """
    Base class for category checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsCategoryBase):
            return CheckOrRelation(self, other)
        
        category_ids = {*self._iter_category_ids(), *other._iter_category_ids()}
        
        return CheckIsAnyCategory(*category_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsCategoryBase):
            return CheckAndRelation(self, other)
        
        category_ids = {*self._iter_category_ids()}&{*other._iter_category_ids()}
        
        return CheckIsAnyCategory(*category_ids)
    
    def _iter_category_ids(self):
        """
        Iterates the category ids of the check.
        
        This method is a generator.
        
        Yields
        ------
        category_id : `int`
        """
        return
        yield


class CheckIsCategory(CheckIsCategoryBase):
    """
    Checks whether the message was sent into any channel of the given category.
    
    Attributes
    ----------
    category_id : `int`
        The respective category's id.
    """
    __slots__ = ('category_id', )
    
    def __new__(cls, category):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        category : `str`, `int`, ``Channel``, ``Guild``
            The category, within sent messages pass the check.
            
            If you want to check whether the channel is not in a category, pass the parameter as the respective guild
            instead.
        
        Raises
        ------
        TypeError
            If `category` was not given neither as ``Channel``, ``Guild``, `str`, `int`.
        ValueError
            If `category` was given as `str`, `int`, but not as a valid snowflake.
        """
        category_id = instance_or_id_to_snowflake(category, (Channel, Guild), 'category')
        
        self = object.__new__(cls)
        self.category_id = category_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        channel = context.message.channel
        
        parent = channel.parent
        if parent is None:
            guild = channel.guild
            if guild is None:
                return False
            
            parent_id = guild.id
        else:
            parent_id = parent.id
        
        if parent_id == self.category_id:
            return True
        
        return False
    
    @copy_docs(CheckIsCategoryBase._iter_category_ids)
    def _iter_category_ids(self):
        yield self.category_id


class CheckIsAnyCategory(CheckIsCategoryBase):
    """
    Checks whether the message was sent into any channel of the given categories.
    
    Attributes
    ----------
    category_ids : `set` of `int`
        The respective categories' id.
    """
    __slots__ = ('category_ids', )
    
    def __new__(cls, *categories):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *categories : `str`, `int`, ``Channel``, ``Guild``
            The categories, within sent messages pass the check.
            
            If you want to check whether the channel is not in a category, pass the parameter as the respective guild
            instead.
        
        Raises
        ------
        TypeError
            If a category was not given neither as ``Channel``, ``Guild``, `str`, `int`.
        ValueError
            If a category was given as `str`, `int`, but not as a valid snowflake.
        """
        category_ids_processed = set()
        
        for category in categories:
            category_id = instance_or_id_to_snowflake(category, (Channel, Guild), 'category')
            category_ids_processed.add(category_id)
        
        category_ids_processed_length = len(category_ids_processed)
        if category_ids_processed_length == 0:
            return CheckBase()
        if category_ids_processed_length == 1:
            return CheckIsCategory(category_ids_processed.pop())
        
        self = object.__new__(cls)
        self.category_ids = category_ids_processed
        return self

    @copy_docs(CheckIsCategoryBase._iter_category_ids)
    def _iter_category_ids(self):
        yield from self.category_ids


class CheckReleaseAt(CheckBase):
    """
    Checks whether the command is already released.
    
    Attributes
    ----------
    release_at : `int`
        The time in snowflake, when the command will be released.
    pre_access_roles : `None`, `set` of ``Role``
        The roles, who are bypassed by the check.
    """
    def __new__(cls, release_at, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        release_at : `datetime`
            When the command is released.
        *roles : `str`, `int`, ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If an element of role was not given neither as ``Role``, `str`, `int`.
        ValueError
            If a role was given as `str`, `int`, but not as a valid snowflake, so a ``Role``
            cannot be precreated with it.
        """
        if not isinstance(release_at, datetime):
            raise TypeError(
                f'`release_at` can be `datetime`, got {release_at.__class__.__name__}; {release_at!r}.'
            )
        
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        release_at = datetime_to_id(release_at)
        
        self = object.__new__(cls)
        self.release_at = release_at
        self.pre_access_roles = roles_processed
        return self
    
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        if message.id > self.release_at:
            return True
        
        user = message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        if context.client.is_owner(user):
            return True
        
        return False


has_role = partial_func(CommandCheckWrapper, CheckHasRole)
owner_or_has_role = partial_func(CommandCheckWrapper, CheckHasRoleOrIsOwner)
has_any_role = partial_func(CommandCheckWrapper, CheckHasAnyRole)
owner_or_has_any_role = partial_func(CommandCheckWrapper, HasAnyRoleCheckOrRelationIsOwner)
guild_only = partial_func(CommandCheckWrapper, CheckIsInGuild)
private_only = partial_func(CommandCheckWrapper, CheckIsInPrivate)
owner_only = partial_func(CommandCheckWrapper, CheckIsOwner)
guild_owner_only = partial_func(CommandCheckWrapper, CheckIsGuildOwner)
owner_or_guild_owner_only = partial_func(CommandCheckWrapper, CheckIsGuildOwnerOrIsOwner)
has_permissions = partial_func(CommandCheckWrapper, CheckHasPermission)
owner_or_has_permissions = partial_func(CommandCheckWrapper, CheckHasPermissionOrIsOwner)
has_guild_permissions = partial_func(CommandCheckWrapper, CheckHasGuildPermission)
owner_or_has_guild_permissions = partial_func(CommandCheckWrapper, CheckHasGuildPermissionOrIsOwner)
has_client_permissions = partial_func(CommandCheckWrapper, CheckHasClientPermission)
has_client_guild_permissions = partial_func(CommandCheckWrapper, CheckHasClientGuildPermission)
is_guild = partial_func(CommandCheckWrapper, CheckIsGuild)
is_any_guild = partial_func(CommandCheckWrapper, CheckIsAnyGuild)
custom = partial_func(CommandCheckWrapper, CheckCustom)
is_channel = partial_func(CommandCheckWrapper, CheckIsChannel)
is_any_channel = partial_func(CommandCheckWrapper, CheckIsAnyChannel)
nsfw_channel_only = partial_func(CommandCheckWrapper, CheckIsNsfwChannel)
announcement_channel_only = partial_func(CommandCheckWrapper, CheckIsAnnouncementChannel)
is_in_voice = partial_func(CommandCheckWrapper, CheckIsInVoice)
booster_only = partial_func(CommandCheckWrapper, CheckIsBooster)
client_only = partial_func(CommandCheckWrapper, CheckIsClient)
user_account_only = partial_func(CommandCheckWrapper, CheckUserAccount)
bot_account_only = partial_func(CommandCheckWrapper, CheckBotAccount)
user_account_or_client_only = partial_func(CommandCheckWrapper, CheckIsUserAccountOrIsClient)
is_category = partial_func(CommandCheckWrapper, CheckIsCategory)
is_any_category = partial_func(CommandCheckWrapper, CheckIsAnyCategory)
release_at = partial_func(CommandCheckWrapper, CheckReleaseAt)



__all__ = ()

import warnings

from scarletio import RichAttributeErrorBaseType


class EmojiCounts(RichAttributeErrorBaseType):
    """
    Represents a guild's counted emojis by their type.
    
    Attributes
    ----------
    managed_animated : `int`
        Animated managed (by an outer integration) emoji count.
    managed_static : `int`
        Static managed (by an outer integration)  emoji count.
    normal_animated : `int`
        Animated emoji count of the guild (excludes managed and premium emojis).
    normal_static : `int`
        Static emoji count of the guild (excludes managed and premium emojis).
    premium_animated : `int`
        Animated premium emoji counts.
    premium_static : `int`
        Static premium emoji counts.
    
    Notes
    -----
    Use ``Guild.emoji_limit`` to get it's emoji limit. That emoji limit is applicable for normal emojis only.
    Premium emojis have their own limit of `25`.
    """
    __slots__ = (
        'managed_animated', 'managed_static', 'normal_animated', 'normal_static', 'premium_animated', 'premium_static'
    )
    
    def __new__(
        cls,
        *,
        managed_animated = 0,
        managed_static = 0,
        normal_animated = 0,
        normal_static = 0,
        premium_animated = 0,
        premium_static = 0,
    ):
        """
        Creates a new emoji counts with the given fields.
        
        Parameters
        ----------
        managed_animated : `int` = `0`, Optional (Keyword only)
            Animated managed (by an outer integration) emoji count.
        managed_static : `int` = `0`, Optional (Keyword only)
            Static managed (by an outer integration)  emoji count.
        normal_animated : `int` = `0`, Optional (Keyword only)
            Animated emoji count of the guild (excludes managed and premium emojis).
        normal_static : `int` = `0`, Optional (Keyword only)
            Static emoji count of the guild (excludes managed and premium emojis).
        premium_animated : `int` = `0`, Optional (Keyword only)
            Animated premium emoji counts.
        premium_static : `int` = `0`, Optional (Keyword only)
            Static premium emoji counts.
        """
        self = object.__new__(cls)
        self.managed_animated = managed_animated
        self.managed_static = managed_static
        self.normal_animated = normal_animated
        self.normal_static = normal_static
        self.premium_animated = premium_animated
        self.premium_static = premium_static
        return self
    
    
    @classmethod
    def from_emojis(cls, iterator):
        """
        Creates a new emoji counts instance from the iterable.
        
        Parameters
        ----------
        iterator : `iterable` of ``Emoji``
            Emoji iterator.
        
        Returns
        -------
        self : `instance<cls>`
        """
        normal_animated = 0
        normal_static = 0
        managed_animated = 0
        managed_static = 0
        premium_static = 0
        premium_animated = 0
        
        for emoji in iterator:
            if emoji.is_premium():
                if emoji.animated:
                    premium_animated += 1
                else:
                    premium_static += 1
                continue
            
            if emoji.managed:
                if emoji.animated:
                    managed_animated += 1
                else:
                    managed_static += 1
                continue
            
            if emoji.animated:
                normal_animated += 1
            else:
                normal_static += 1
            continue
        
        
        self = object.__new__(cls)
        self.managed_animated = managed_animated
        self.managed_static = managed_static
        self.normal_animated = normal_animated
        self.normal_static = normal_static
        self.premium_animated = premium_animated
        self.premium_static = premium_static
        return self
    
    
    @property
    def managed_total(self):
        """
        Returns the total amount of managed emojis.
        
        Returns
        -------
        managed_total : `int`
        """
        return self.managed_animated + self.managed_static
    
    
    @property
    def normal_total(self):
        """
        Returns the total amount of normal emojis.
        
        Returns
        -------
        normal_total : `int`
        """
        return self.normal_animated + self.normal_static
    
    
    @property
    def premium_total(self):
        """
        Returns the total amount of premium emojis.
        
        Returns
        -------
        premium_total : `int`
        """
        return self.premium_animated + self.premium_static
    
    
    @property
    def animated_total(self):
        """
        Returns the total amount of animated emojis.
        
        Returns
        -------
        animated_total : `int`
        """
        return self.managed_animated + self.normal_animated + self.premium_animated
    
    
    @property
    def static_total(self):
        """
        Returns the total amount static emojis.
        
        Returns
        -------
        static_total : `int`
        """
        return self.managed_static + self.normal_static + self.premium_static
    
    
    @property
    def total(self):
        """
        Returns the total count of emojis.
        
        Returns
        -------
        total : `int`
        """
        return (
            self.managed_animated +
            self.managed_static +
            self.normal_animated +
            self.normal_static +
            self.premium_animated +
            self.premium_static
        )
    
    
    def _iter_field_values(self):
        """
        Iterates over the fields of the emoji counts.
        
        This method is an iterable generator.
        
        Yields
        ------
        field_value : `int`
        """
        yield self.managed_animated
        yield self.managed_static
        yield self.normal_animated
        yield self.normal_static
        yield self.premium_animated
        yield self.premium_static
    
    
    def __bool__(self):
        """Returns whether the emoji counts counted any emojis"""
        for field_value in self._iter_field_values():
            if field_value:
                return True
        
        return False
    
    
    def __repr__(self):
        """Returns the emoji counts' representation"""
        repr_parts = [self.__class__.__name__, '(']
        field_added = False
        
        for field_name, field_value in zip(self.__slots__, self._iter_field_values()):
            if field_value:
                if field_added:
                    repr_parts.append(', ')
                else:
                    field_added = True
                
                repr_parts.append(field_name)
                repr_parts.append(' = ')
                repr_parts.append(repr(field_value))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two emoji counts equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        for self_field, other_field in zip(self._iter_field_values(), other._iter_field_values()):
            if self_field != other_field:
                return False
        
        return True
    
    
    def __hash__(self):
        """Returns the emoji counts' hash value."""
        hash_value = 0
        
        shift = 0
        
        for field_value in self._iter_field_values():
            hash_value ^= field_value << shift
            shift += 4
        
        return hash_value
    
    
    def __iter__(self):
        """Deprecated and will be removed in 2023 April."""
        warnings.warn(
            (
                f'Unpacking `Guild.emoji_counts` is deprecated and will be removed in 2023 April.\n'
                f'Please do the following instead:\n'
                f'emoji_counts = guild.emoji_counts\n'
                f'normal_static = emoji_counts.normal_static\n'
                f'normal_animated = emoji_counts.normal_animated\n'
                f'managed_static = emoji_counts.managed_static\n'
                f'managed_animated = emoji_counts.managed_animated'
            ),
            FutureWarning,
            stacklevel = 2,
        )

        yield self.normal_static
        yield self.normal_animated
        yield self.managed_static
        yield self.managed_animated



__all__ = ()

from .....discord.component.shared_constants import CUSTOM_ID_LENGTH_MAX

from ...converters import (
    RegexMatcher, check_component_converters_satisfy_regex, check_component_converters_satisfy_string
)

try:
    # CPython
    from re import Pattern
except ImportError:
    # ChadPython (PyPy)
    from re import _pattern_type as Pattern


def _validate_custom_id(custom_id):
    """
    Validates a `custom_id` value.
    
    Parameters
    ----------
    custom_id : `str`, `re.Pattern`
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `str`, `re.Pattern`
        The validated custom_id.
    
    Raises
    ------
    ValueError
        If `custom_id`'s length is out of the expected range.
    """
    if isinstance(custom_id, str):
        if type(custom_id) is not str:
            custom_id = str(custom_id)
        
        custom_id_length = len(custom_id)
        if (custom_id_length < 1) or (custom_id_length > CUSTOM_ID_LENGTH_MAX):
            raise ValueError(
                f'`custom_id` length can be in range [1:{CUSTOM_ID_LENGTH_MAX}], got '
                f'{custom_id_length}; {custom_id!r}.'
            )
    
    return custom_id


def _validate_custom_ids(custom_id):
    """
    Validates one or more `custom_id` values.
    
    Parameters
    ----------
    custom_id : `str`, (`list`, `set`) of `str`.
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `set` of (`str`, `re.Pattern`)
        The non-duped custom-ids.
    
    Raises
    ------
    TypeError
        If `custom_id`'s type is incorrect.
    ValueError
        If a `custom_id`'s length is out of the expected range.
    """
    custom_ids = set()
    if isinstance(custom_id, (str, Pattern)):
        custom_id = _validate_custom_id(custom_id)
        custom_ids.add(custom_id)
    
    elif isinstance(custom_id, (list, set)):
        
        if not custom_id:
            raise ValueError(
                f'`custom_id` received as empty {custom_id.__class__.__name__}.'
            )
        
        for sub_custom_id in custom_id:
            if isinstance(sub_custom_id, (str, Pattern)):
                sub_custom_id = _validate_custom_id(sub_custom_id)
                custom_ids.add(sub_custom_id)
                continue
            
            raise TypeError(
                f'`custom_id` contains a non `str` element, got: '
                f'{sub_custom_id.__class__.__name__}; {sub_custom_id!r}; custom_id = {custom_id!r}.'
            )
    
    else:
        raise TypeError(
            f'`custom_id` can be `str`, (`list`, `set`) of `str`, got '
            f'{custom_id.__class__.__name__}; {custom_id!r}.'
        )
    
    return custom_ids


def _validate_name(name):
    """
    Validates the given name.
    
    Parameters
    ----------
    name : `None`, `str`
        A command's respective name.
    
    Returns
    -------
    name : `None`, `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` is not given as `None` neither as `str`.
    """
    if name is not None:
        name_type = name.__class__
        if name_type is str:
            pass
        elif issubclass(name_type, str):
            name = str(name)
        else:
            raise TypeError(
                f'`name` can be `None`, `str`, got {name_type.__name__}; {name!r}.'
            )
    
    return name


def split_and_check_satisfaction(custom_ids, parameter_converters):
    """
    Splits custom id-s to `str` and to `re.Pattern`-s and validates them.
    
    Parameters
    ----------
    custom_ids : `set` of (`str`, `re.Pattern`)
        The custom-ids to split and validate.
    parameter_converters : `tuple` of ``ParameterConverter``
        The parameter converters generated from a component command.
    
    Returns
    -------
    string_custom_ids : `None`, `tuple` of `str`
        String custom ids.
    regex_custom_ids : `None`, `tuple` of ``RegexMatcher``
        Regex custom ids.
    
    Raises
    ------
    ValueError
        A string or regex pattern is not satisfied.
    """
    # Build
    string_custom_ids = None
    regex_custom_ids = None
    for custom_id in custom_ids:
        if isinstance(custom_id, str):
            if string_custom_ids is None:
                string_custom_ids = []
            
            string_custom_ids.append(custom_id)
        else:
            if regex_custom_ids is None:
                regex_custom_ids = []
            
            regex_custom_ids.append(custom_id)
    
    # Convert
    if (string_custom_ids is not None):
        string_custom_ids = tuple(string_custom_ids)
    
    if (regex_custom_ids is not None):
        regex_custom_ids = tuple(RegexMatcher(regex_custom_id) for regex_custom_id in regex_custom_ids)
    
    # Check
    if (string_custom_ids is not None):
        check_component_converters_satisfy_string(parameter_converters)
    
    if (regex_custom_ids is not None):
        for regex_custom_id in regex_custom_ids:
            check_component_converters_satisfy_regex(parameter_converters, regex_custom_id)
    
    # Good
    return string_custom_ids, regex_custom_ids



__all__ = ()

from .preinstanced import Locale


LOCALE_DEFAULT = Locale.english_us
Locale.INSTANCES[''] = LOCALE_DEFAULT


def build_locale_dictionary(dictionary):
    """
    Builds a locale dictionary where they keys are all ``Locale``-s.
    
    Parameters
    ----------
    dictionary : `dict` of (`str`, `object`) items
        The dictionary to process.
    
    Returns
    -------
    transformed : `dict` of (``Locale``, `object`) items
    """
    if (dictionary is not None) and dictionary:
        return {Locale.get(key): value for key, value in dictionary.items()}


def destroy_locale_dictionary(dictionary):
    """
    Builds a json serializable dictionary where they keys are all `str`-s.
    
    Parameters
    ----------
    dictionary : `dict` of (`Locale`, `object`) items
        The dictionary to process.
    
    Returns
    -------
    transformed : `dict` of (`str`, `object`) items
    """
    if dictionary is not None:
        return {key.value: value for key, value in dictionary.items()}


def hash_locale_dictionary(dictionary):
    """
    Hashes a locale dictionary where they keys are all ``Locale``-s.
    
    Parameters
    ----------
    dictionary : `dict` of (`str`, `object`) items
        The dictionary to process.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = 0
    
    for key, value in dictionary.items():
        hash_value ^= hash(key.value) & hash(value)
    
    return hash_value



__all__ = ('Command',)

from scarletio import RichAttributeErrorBaseType, WeakReferer

from ..constants import REGISTERED_COMMANDS, REGISTERED_COMMANDS_BY_NAME

from .helpers import normalize_alters, normalize_command_name
from .category import CommandCategory
from .result import COMMAND_RESULT_CODE_COMMAND_UNINITIALIZED, CommandResult


class Command(RichAttributeErrorBaseType):
    """
    Represents a command line command.
    
    Attributes
    ----------
    _command_category : `None`, ``CommandCategory``
        Command category of the command.
    _self_reference : `None`, ``WeakReferer``
        Reference to itself.
    alters : `None`, `set` of `str`
        Alternative names for the command.
    name : `str`
        The command's name.
    """
    __slots__ = ('__weakref__', '_command_category', '_self_reference', 'alters', 'name')
    
    def __new__(cls, name, alters):
        """
        Creates a new command line command.
        
        Parameters
        ----------
        name : `str`
            The command's name.
        alters : `None`, `str`, `iterable` of `str`
            Alternative names for the command.
        
        Raises
        ------
        TypeError
            Type of `alters` is unaccepted.
        """
        name = normalize_command_name(name)
        alters = normalize_alters(alters, name)
        
        self = object.__new__(cls)
        self._command_category = None
        self.alters = alters
        self.name = name
        self._self_reference = None
        
        self._self_reference = WeakReferer(self)
        self._command_category = CommandCategory(self, None)
        
        REGISTERED_COMMANDS.add(self)
        
        names = [self.name]
        alters = self.alters
        if (alters is not None):
            names.extend(alters)
        
        for name_ in names:
            
            try:
                already_registered_command = REGISTERED_COMMANDS_BY_NAME[name_]
            except KeyError:
                pass
            else:
                already_registered_command._unregister_name(name_)
            
            REGISTERED_COMMANDS_BY_NAME[name_] = self
        
        return self
    
    
    def __repr__(self):
        """Returns the command's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        command_category = self._command_category
        if (command_category is not None):
            repr_parts.append(', command_category = ')
            repr_parts.append(repr(command_category))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def register_command_category(self, name):
        """
        Registers a sub command to the command.
        
        Parameters
        ----------
        name : `str`
            The name of the sub-command.
        
        Returns
        -------
        sub_command : ``CommandCategory``
        """
        return self._command_category.register_command_category(name)
    
    
    def register_command_function(self, function, name, description):
        """
        Parameters
        ----------
        function : `callable`
            The function to call when the command is used.
        name : `str`
            The command's name.
        description : `None`, `str`
            The command's description.
        
        Returns
        -------
        command_function : ``CommandFunction``
        """
        return self._command_category.register_command_function(function, name, description)
    
    
    def invoke(self, parameters, index):
        """
        Calls the command line command.
        
        Parameters
        ----------
        parameters : `list` of `str`
            Command line parameters.
        index : `int`
            The index of the first parameter trying to process.
        
        Returns
        -------
        command_result : ``CommandResult``
        """
        command_category = self._command_category
        if (command_category is None):
            return CommandResult(
                COMMAND_RESULT_CODE_COMMAND_UNINITIALIZED,
                self,
            )
        else:
            return command_category.invoke(parameters, index)
    
    
    def _trace_back_name(self):
        """
        Traces back to the source name of the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
        
        Returns
        -------
        name : `str`
        """
        name = self.name
        yield name
        return name
    
    
    def walk_usage(self):
        """
        Walks over the usage of the command.
        
        This method is an iterable generator.
        
        Yields
        -------
        usage : `str`
        """
        for into in self.walk_usage_into([]):
            yield ''.join(into)
            into.clear()
    
    
    def get_direct_usage(self, *sub_command_stack):
        """
        Returns the direct usage of the command for the given sub-command stack.
        
        Parameters
        ----------
        *sub_command_stack : `str`
            Sub command stack to get the direct usage for.
        
        Returns
        -------
        usage : `str`
        """
        return ''.join(self.render_direct_usage_into([], *sub_command_stack))
    
    
    def render_direct_usage_into(self, into, *sub_command_stack):
        """
        Renders the direct usage of the command for the given sub-command stack.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the usage into.
        *sub_command_stack : `str`
            Sub command stack to get the direct usage for.
        
        Returns
        -------
        into : `list` of `str`
        """
        return self._command_category.render_direct_usage_into(into, *sub_command_stack)
    
    
    def walk_usage_into(self, into):
        """
        Walks over the usage of the command and renders it to the given list.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the commands into.
        
        Yields
        -------
        into : `list` of `str`
        
        Returns
        -------
        into : `list` of `str`
        """
        return (yield from self._command_category.walk_usage_into(into))
    
    
    def iter_command_functions(self):
        """
        Iterates over the command functions of the command category.
        
        This method is an iterable generator.
        
        Yields
        ------
        command_function : ``CommandFunction``
        """
        yield from self._command_category.iter_command_functions()
    
    
    def _unregister_name(self, name):
        """
        Unregisters the command for the given name.
        
        Parameters
        ----------
        name : `str`
            The name to unregister.
        """
        if self.name == name:
            self._unregister()
        else:
            self._unregister_alter(name)
    
    
    def _unregister(self):
        """
        Unregisters the command. Fully.
        """
        REGISTERED_COMMANDS.discard(self)
        
        names = [self.name]
        alters = self.alters
        if (alters is not None):
            names.extend(alters)
        
        for name in names:
            if REGISTERED_COMMANDS_BY_NAME.get(name) is self:
                del REGISTERED_COMMANDS_BY_NAME[name]
    
    
    def _unregister_alter(self, name):
        """
        Removes an alternative name of the command.
        
        Parameters
        ----------
        name : `str`
            The alternative name to unregister.
        """
        if REGISTERED_COMMANDS_BY_NAME.get(name) is self:
            del REGISTERED_COMMANDS_BY_NAME[name]
        
        alters = self.alters
        try:
            alters.remove(name)
        except KeyError:
            pass
        else:
            if not alters:
                self.alters = None



from .constants import *
from .fields import *
from .utils import *
from .verification_screen import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *utils.__all__,
    *verification_screen.__all__,
)



__all__ = ('MessageInteraction',)

from scarletio import include

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ZEROUSER

from .fields import (
    parse_id, parse_name_and_sub_command_name_stack, parse_type, parse_user, put_id_into,
    put_name_and_sub_command_name_stack_into, put_type_into, put_user_into, validate_id, validate_name,
    validate_sub_command_name_stack, validate_type, validate_user
)


InteractionType = include('InteractionType')


PRECREATE_FIELDS = {
    'name': ('name', validate_name),
    'sub_command_name_stack': ('sub_command_name_stack', validate_sub_command_name_stack),
    'interaction_type': ('type', validate_type),
    'user': ('user', validate_user),
}

    
class MessageInteraction(DiscordEntity):
    """
    Sent with a ``Message`` when the it is a response to an ``InteractionEvent``.
    
    Attributes
    ----------
    id : `int`
        The interaction's identifier.
    name : `str`
        The invoked interaction's name.
    sub_command_name_stack : `None`, `tuple` of `str`
        The sub-command-group and sub-command names.
    type : ``InteractionType``
        The interaction's type.
    user : ``ClientUserBase``
        Who invoked the interaction.
    """
    __slots__ = ('name', 'sub_command_name_stack', 'type', 'user')
    
    def __new__(cls, *, interaction_type = ..., name = ..., sub_command_name_stack = ..., user = ...):
        """
        Creates a new message interaction with the given fields.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            Who invoked the interaction.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # interaction_type
        if interaction_type is ...:
            interaction_type = InteractionType.none
        else:
            interaction_type = validate_type(interaction_type)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # sub_command_name_stack
        if sub_command_name_stack is ...:
            sub_command_name_stack = None
        else:
            sub_command_name_stack = validate_sub_command_name_stack(sub_command_name_stack)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.id = 0
        self.name = name
        self.sub_command_name_stack = sub_command_name_stack
        self.type = interaction_type
        self.user = user
        return self
    
    
    @classmethod
    def precreate(cls, message_interaction_id, **keyword_parameters):
        """
        Creates a new message interaction with the given predefined fields.
        
        > Since message interactions are not globally cached, this method is only used for testing.
        
        Parameters
        ----------
        message_interaction_id : `int`
            The message interaction's id.
        
        **keyword_parameters : Keyword parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            Who invoked the interaction.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        message_interaction_id = validate_id(message_interaction_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(message_interaction_id)
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, message_interaction_id):
        """
        Creates a new message interaction with it's defaults attributes set.
        
        Parameters
        ----------
        message_interaction_id : `int`
            The message interaction's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = message_interaction_id
        self.name = ''
        self.sub_command_name_stack = None
        self.type = InteractionType.none
        self.user = ZEROUSER
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new message interaction from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Message interaction data.
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective message's guild's identifier.
        
        Returns
        -------
        data : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.id = parse_id(data)
        self.name, self.sub_command_name_stack = parse_name_and_sub_command_name_stack(data)
        self.type = parse_type(data)
        self.user = parse_user(data, guild_id)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False, guild_id = 0):
        """
        Tries to convert the message interaction back to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective message's guild's identifier.
        
        Returns
        -------
        data : `dict` of (`str`, `object`)
        """
        data = {}
        put_name_and_sub_command_name_stack_into((self.name, self.sub_command_name_stack), data, defaults)
        put_type_into(self.type, data, defaults)
        
        if include_internals:
            put_id_into(self.id, data, defaults)
            put_user_into(self.user, data, defaults, guild_id = guild_id)
        
        return data
    
    
    @property
    def partial(self):
        """
        Returns whether the message interaction is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    
    def __repr__(self):
        """Returns the message interaction's representation."""
        repr_parts = ['<', self.__class__.__name__, ' id = ', repr(self.id), ', type = ']
        
        interaction_type = self.type
        repr_parts.append(interaction_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(interaction_type.value))
        
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            repr_parts.append(', sub_command_name_stack = ')
            repr_parts.append(repr(sub_command_name_stack))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two message interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two message interactions not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id and self_id != other_id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # sub_command_name_stack
        if self.sub_command_name_stack != other.sub_command_name_stack:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # user
        if self.user != other.user:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the message integration's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # sub_command_name_stack
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            hash_value ^= len(sub_command_name_stack)
            for sub_command_name in sub_command_name_stack:
                hash_value ^= hash(sub_command_name)
        
        # type
        hash_value ^= self.type.value << 4
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    
    def copy(self):
        """
        Copies the message application returning a new partial one.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        
        new.id = 0
        new.name = self.name
        
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            sub_command_name_stack = (*sub_command_name_stack,)
        new.sub_command_name_stack = sub_command_name_stack
        
        new.type = self.type
        new.user = self.user
        
        return new
    
    
    def copy_with(self, *, interaction_type = ..., name = ..., sub_command_name_stack = ..., user = ...):
        """
        Copies the message interaction with the given fields returning a new partial one.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            Who invoked the interaction.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # interaction_type
        if interaction_type is ...:
            interaction_type = self.type
        else:
            interaction_type = validate_type(interaction_type)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # sub_command_name_stack
        if sub_command_name_stack is ...:
            sub_command_name_stack = self.sub_command_name_stack
            if (sub_command_name_stack is not None):
                sub_command_name_stack = (*sub_command_name_stack,)
        else:
            sub_command_name_stack = validate_sub_command_name_stack(sub_command_name_stack)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
    
        # Construct
        self = object.__new__(type(self))
        self.id = 0
        self.name = name
        self.sub_command_name_stack = sub_command_name_stack
        self.type = interaction_type
        self.user = user
        return self
    
    
    @property
    def joined_name(self):
        """
        Returns the joined name of the message interaction.
        
        Returns
        -------
        joined_name : `str`
        """
        name = self.name
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is None):
            return name
        
        return ' '.join([name, *sub_command_name_stack])
    
    



__all__ = ()

from ...application import Application
from ...field_parsers import entity_id_parser_factory, nullable_entity_array_parser_factory
from ...field_putters import entity_id_putter_factory, nullable_entity_array_putter_factory
from ...field_validators import entity_id_validator_factory, nullable_entity_array_validator_factory
from ...guild import Guild

from ..application_command import ApplicationCommand
from ..application_command_permission_overwrite import ApplicationCommandPermissionOverwrite

# application_command_id

parse_application_command_id = entity_id_parser_factory('id')
put_application_command_id_into = entity_id_putter_factory('id')
validate_application_command_id = entity_id_validator_factory(
    'application_command_id', ApplicationCommand
)

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id_into = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# permission_overwrites

parse_permission_overwrites = nullable_entity_array_parser_factory('permissions', ApplicationCommandPermissionOverwrite)
put_permission_overwrites_into = nullable_entity_array_putter_factory(
    'permissions', ApplicationCommandPermissionOverwrite
)
validate_permission_overwrites = nullable_entity_array_validator_factory(
    'permission_overwrites', ApplicationCommandPermissionOverwrite
)



__all__ = ()

from ...activity import Activity
from ...channel import Channel
from ...field_parsers import default_entity_parser_factory, entity_id_parser_factory
from ...field_putters import entity_id_optional_putter_factory, entity_id_putter_factory
from ...field_validators import default_entity_validator, entity_id_set_validator_factory, entity_id_validator_factory
from ...user import ClientUserBase

from .constants import ACTIVITY_KEY

# activity

parse_activity = default_entity_parser_factory(ACTIVITY_KEY, Activity, default_factory = lambda : Activity())


def put_activity_into(activity, data, defaults):
    """
    Puts the activity into the given data.
    
    Parameters
    ----------
    activity : ``Activity``
        The activity to put into the given data.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data[ACTIVITY_KEY] = activity.to_data(include_internals = True, user = True)
    return data


validate_activity = default_entity_validator('activity', Activity, default_factory = lambda : Activity())

# application_id

def parse_application_id(data):
    """
    Parses application identifier from the given embedded activity state data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Embedded activity state data.
    
    Returns
    -------
    application_application_id : `int`
    """
    activity_data = data.get(ACTIVITY_KEY, None)
    if activity_data is None:
        return 0
    
    application_id = activity_data.get('application_id', None)
    if application_id is None:
        return 0
    
    return int(application_id)


def put_application_id_into(application_id, data, defaults):
    """
    Puts the application identifier into the given data.
    
    Parameters
    ----------
    application_id : `int`
        The represented application's identifier.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if application_id or defaults:
        activity_data = data.get(ACTIVITY_KEY, None)
        if activity_data is None:
            activity_data = {}
            data[ACTIVITY_KEY] = activity_data
        
        if application_id:
            application_id = str(application_id)
        else:
            application_id = None
        
        activity_data['application_id'] = application_id
        
    return data


validate_application_id = entity_id_validator_factory('application_id', NotImplemented, include = 'Application')

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# user_ids

def parse_user_ids(data):
    """
    Parses the user identifiers from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Embedded activity state data.
    
    Returns
    -------
    user_ids : `set` of `int`
    """
    user_ids = data.get('users', None)
    if user_ids is None:
        return set()
    
    return {int(user_id) for user_id in user_ids}


def put_user_ids_into(user_ids, data, defaults):
    """
    Puts the user identifiers into the given data.
    
    Parameters
    ----------
    user_ids : `set` of `int`
        The user identifiers to put into the given data.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['users'] = [str(user_id) for user_id in user_ids]
    return data


validate_user_ids = entity_id_set_validator_factory('user_ids', ClientUserBase)



__all__ = ('OnboardingPromptType',)

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase


@export
class OnboardingPromptType(PreinstancedBase):
    """
    The type of an onboarding prompt.
    
    Attributes
    ----------
    name : `str`
        The name of the onboarding prompt type.
    value : `int`
        The identifier value the onboarding prompt type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``OnboardingPromptType``) items
        Stores the predefined ``OnboardingPromptType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The application command option types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the onboarding prompt types.
    
    Every predefined onboarding prompt type can be accessed as class attribute as well:
    
    +-----------------------------------+-----------------------------------+-------+
    | Class attribute name              | Name                              | Value |
    +===================================+===================================+=======+
    | multiple_choice                   | multiple choice                   | 0     |
    +-----------------------------------+-----------------------------------+-------+
    | dropdown                          | dropdown                          | 1     |
    +-----------------------------------+-----------------------------------+-------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    
    multiple_choice = P(0, 'multiple choice')
    dropdown = P(1, 'dropdown')



__all__ = ()

from scarletio import IgnoreCaseString
from scarletio.web_common.headers import AUTHORIZATION as HEADER_AUTHORIZATION


LAVALINK_KEY_STATS_UPTIME = 'uptime'
LAVALINK_KEY_STATS_PLAYER_COUNT = 'players'
LAVALINK_KEY_STATS_PLAYING_PLAYER_COUNT = 'playingPlayers'

LAVALINK_KEY_STATS_MEMORY = 'memory'
LAVALINK_KEY_STATS_MEMORY_FREE = 'free'
LAVALINK_KEY_STATS_MEMORY_USED = 'used'
LAVALINK_KEY_STATS_MEMORY_ALLOCATED = 'allocated'
LAVALINK_KEY_STATS_MEMORY_RESERVABLE = 'reservable'

LAVALINK_KEY_STATS_CPU = 'cpu'
LAVALINK_KEY_STATS_CPU_CORES = 'cores'
LAVALINK_KEY_STATS_CPU_SYSTEM_LOAD = 'systemLoad'
LAVALINK_KEY_STATS_CPU_LAVALINK_LOAD = 'lavalinkLoad'

LAVALINK_KEY_STATS_FRAME = 'frameStats'
LAVALINK_KEY_STATS_FRAME_SENT = 'sent'
LAVALINK_KEY_STATS_FRAME_NULLED = 'nulled'
LAVALINK_KEY_STATS_FRAME_DEFICIT = 'deficit'


LAVALINK_KEY_THRESHOLD_MS = 'thresholdMs'
LAVALINK_KEY_GUILD_ID = 'guildId'
LAVALINK_KEY_SESSION_ID = 'sessionId'
LAVALINK_KEY_VOICE_SERVER_UPDATE_EVENT = 'event'
LAVALINK_KEY_PAUSE = 'pause'
LAVALINK_KEY_POSITION_MS = 'position'
LAVALINK_KEY_START_TIME = 'startTime'
LAVALINK_KEY_END_TIME = 'endTime'
LAVALINK_KEY_NO_REPLACE = 'noReplace'
LAVALINK_KEY_TRACK = 'track'
LAVALINK_KEY_PLAYER_STATE = 'state'
LAVALINK_KEY_EVENT_TYPE = 'type'
LAVALINK_KEY_END_REASON = 'reason'
LAVALINK_KEY_EXCEPTION_REASON_DEPRECATED = 'error'
LAVALINK_KEY_EXCEPTION_REASON_OBJECT = 'exception'
LAVALINK_KEY_EXCEPTION_REASON_OBJECT_MESSAGE = 'message'
LAVALINK_KEY_EXCEPTION_REASON_OBJECT_SEVERITY = 'severity'
LAVALINK_KEY_WEBSOCKET_CLOSE_CODE = 'code'
LAVALINK_KEY_WEBSOCKET_CLOSE_REASON = 'reason'
LAVALINK_KEY_WEBSOCKET_CLOSE_BY_REMOTE = 'byRemote'
LAVALINK_KEY_NODE_RESUME_KEY = 'key'
LAVALINK_KEY_TRACKS = 'tracks'
LAVALINK_KEY_PLAYLIST = 'playlistInfo'
LAVALINK_KEY_PLAYLIST_NAME = 'name'
LAVALINK_KEY_PLAYLIST_SELECTED_TRACK_INDEX = 'selectedTrack'


LAVALINK_KEY_FILTER_EQUALIZER = 'equalizer'
LAVALINK_KEY_FILTER_EQUALIZER_BAND = 'band'
LAVALINK_KEY_FILTER_EQUALIZER_GAIN = 'gain'

LAVALINK_KEY_FILTER_KARAOKE = 'karaoke'
LAVALINK_KEY_FILTER_KARAOKE_LEVEL = 'level'
LAVALINK_KEY_FILTER_KARAOKE_MONO_LEVEL = 'monoLevel'
LAVALINK_KEY_FILTER_KARAOKE_FILTER_BAND = 'filterBand'
LAVALINK_KEY_FILTER_KARAOKE_FILTER_WIDTH = 'filterWidth'

LAVALINK_KEY_FILTER_TIMESCALE = 'timescale'
LAVALINK_KEY_FILTER_TIMESCALE_PITCH = 'pitch'
LAVALINK_KEY_FILTER_TIMESCALE_RATE = 'rate'
LAVALINK_KEY_FILTER_TIMESCALE_SPEED = 'speed'


LAVALINK_KEY_FILTER_TREMOLO = 'tremolo'
LAVALINK_KEY_FILTER_TREMOLO_DEPTH = 'depth'
LAVALINK_KEY_FILTER_TREMOLO_FREQUENCY = 'frequency'


LAVALINK_KEY_FILTER_VIBRATO = 'vibrato'
LAVALINK_KEY_FILTER_VIBRATO_DEPTH = 'depth'
LAVALINK_KEY_FILTER_VIBRATO_FREQUENCY = 'frequency'

LAVALINK_KEY_FILTER_ROTATION = 'rotation'
LAVALINK_KEY_FILTER_ROTATION_ROTATION = 'rotationHz'

LAVALINK_KEY_FILTER_DISTORTION = 'distortion'
LAVALINK_KEY_FILTER_DISTORTION_COS_OFFSET = 'cosOffset'
LAVALINK_KEY_FILTER_DISTORTION_COS_SCALE = 'cosScale'
LAVALINK_KEY_FILTER_DISTORTION_SIN_OFFSET = 'sinOffset'
LAVALINK_KEY_FILTER_DISTORTION_SIN_SCALE = 'sinScale'
LAVALINK_KEY_FILTER_DISTORTION_TAN_OFFSET = 'tanOffset'
LAVALINK_KEY_FILTER_DISTORTION_TAN_SCALE = 'tanScale'
LAVALINK_KEY_FILTER_DISTORTION_OFFSET = 'offset'
LAVALINK_KEY_FILTER_DISTORTION_SCALE = 'scale'


LAVALINK_KEY_FILTER_CHANNEL_MIX = 'channelMix'
LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_LEFT = 'leftToLeft'
LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_RIGHT = 'leftToRight'
LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_LEFT = 'rightToLeft'
LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_RIGHT = 'rightToRight'


LAVALINK_KEY_FILTER_LOW_PASS = 'lowPass'
LAVALINK_KEY_FILTER_LOW_PASS_SMOOTHING = 'smoothing'


LAVALINK_KEY_GATEWAY_OPERATION_STATS = 'stats'
LAVALINK_KEY_GATEWAY_OPERATION_PLAYER_UPDATE = 'playerUpdate'
LAVALINK_KEY_GATEWAY_OPERATION_EVENT = 'event'

LAVALINK_KEY_FILTER_VOLUME = 'volume'


LAVALINK_KEY_NODE_OPERATION = 'op'
LAVALINK_KEY_NODE_OPERATION_VOICE_UPDATE = 'voiceUpdate'
LAVALINK_KEY_NODE_OPERATION_PLAYER_DESTROY = 'destroy'
LAVALINK_KEY_NODE_OPERATION_PLAYER_STOP = 'stop'
LAVALINK_KEY_NODE_OPERATION_PLAYER_PAUSE = 'pause'
LAVALINK_KEY_NODE_OPERATION_PLAYER_SEEK = 'seek'
LAVALINK_KEY_NODE_OPERATION_PLAYER_EDIT_BANDS = 'equalizer'
LAVALINK_KEY_NODE_OPERATION_PLAYER_PLAY = 'play'
LAVALINK_KEY_NODE_OPERATION_PLAYER_FILTER = 'filters'
LAVALINK_KEY_NODE_OPERATION_SET_RESUME_KEY = 'configureResuming'

LAVALINK_KEY_EVENT_TRACK_END = 'TrackEndEvent'
LAVALINK_KEY_EVENT_TRACK_EXCEPTION = 'TrackExceptionEvent'
LAVALINK_KEY_EVENT_TRACK_START = 'TrackStartEvent'
LAVALINK_KEY_EVENT_TRACK_STUCK = 'TrackStuckEvent'
LAVALINK_KEY_EVENT_PLAYER_WEBSOCKET_CLOSED = 'WebSocketClosedEvent'


LAVALINK_KEY_TRACK_BASE64 = 'track'
LAVALINK_KEY_TRACK_DICT = 'info'
LAVALINK_KEY_TRACK_IDENTIFIER = 'identifier'
LAVALINK_KEY_TRACK_SEEKABLE = 'isSeekable'
LAVALINK_KEY_TRACK_AUTHOR = 'author'
LAVALINK_KEY_TRACK_DURATION_MS = 'length'
LAVALINK_KEY_TRACK_IS_STREAM = 'isStream'
LAVALINK_KEY_TRACK_TITLE = 'title'
LAVALINK_KEY_TRACK_URL = 'uri'
LAVALINK_KEY_TRACK_POSITION_MS = 'position'


LAVALINK_KEY_ROUTEPLANNER_TYPE = 'class'
LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_IP = 'RotatingIpRoutePlanner'
LAVALINK_KEY_ROUTEPLANNER_TYPE_NANO_IP = 'NanoIpRoutePlanner'
LAVALINK_KEY_ROUTEPLANNER_TYPE_ROTATING_NANO_IP ='RotatingNanoIpRoutePlanner'
LAVALINK_KEY_ROUTEPLANNER_OBJECT = 'details'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK = 'ipBlock'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_TYPE = 'type'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_BLOCK_SIZE_STRING = 'size'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESSES = 'failingAddresses'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_ADDRESS = 'address'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_FAILING_ADDRESS_UNIX_TIME = 'failingTimestamp'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_BLOCK_INDEX_STRING = 'blockIndex'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS_INDEX_STRING = 'currentAddressIndex'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_ROTATE_INDEX_STRING = 'rotateIndex'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_IP_INDEX_STRING = 'ipIndex'
LAVALINK_KEY_ROUTEPLANNER_OBJECT_CURRENT_ADDRESS = 'currentAddress'

HEADER_USER_ID = IgnoreCaseString('User-Id')
HEADER_CLIENT_NAME = IgnoreCaseString('Client-Name')
HEADER_SHARD_COUNT = IgnoreCaseString('Num-Shards')
HEADER_RESUME_KEY = IgnoreCaseString('Resume-Key')


LAVALINK_BAND_COUNT = 15



__all__ = ('mark_as_plugin_root_directory',)

from os import listdir as list_directory
from os.path import (
    basename as get_file_name, dirname as get_directory_name, isdir as is_directory, isfile as is_file,
    join as join_paths
)

from scarletio import get_last_module_frame

from ..constants import IGNORED_DIRECTORY_NAMES

from .import_plugin_ import import_plugin


def mark_as_plugin_root_directory():
    """
    Imports all sub-plugins from the directory.
    
    Returns
    -------
    count : `int`
        The amount of sub-plugins imported.
    
    Raises
    ------
    RuntimeError
        - If called from other file than `__init__.py`.
    """
    module_frame = get_last_module_frame()
    if (module_frame is None):
        return 0
    
    frame_globals = module_frame.f_globals
    spec = frame_globals.get('__spec__', None)
    if spec is None:
        return
    
    source_file_path = spec.origin

    if get_file_name(source_file_path) != '__init__.py':
        raise RuntimeError(
            f'Cannot set directory level environment, top level file is not an `__init__.py` file.'
        )
    
    frame_globals.setdefault('__all__', ())
    
    source_directory_path = get_directory_name(source_file_path)
    source_name = spec.name
    
    for file_name in list_directory(source_directory_path):
        if file_name == '__init__.py':
            continue
        
        file_path = join_paths(source_directory_path, file_name)
        if is_file(file_path):
            if not file_path.endswith('.py'):
                continue
            
            file_name = file_name[:-len('.py')]
        
        elif is_directory(file_path):
            if file_name in IGNORED_DIRECTORY_NAMES:
                continue
            
            if not is_file(join_paths(file_path, '__init__.py')):
                continue
        
        else:
            continue
        
        import_plugin(f'{source_name}.{file_name}')
        continue



__all__ = ('ChannelMetadataGuildTextBase',)

from scarletio import copy_docs

from .constants import AUTO_ARCHIVE_DEFAULT, SLOWMODE_DEFAULT
from .fields import (
    parse_default_thread_auto_archive_after, parse_default_thread_slowmode, parse_nsfw, parse_slowmode, parse_topic,
    put_default_thread_auto_archive_after_into, put_default_thread_slowmode_into, put_nsfw_into, put_slowmode_into,
    put_topic_into, validate_default_thread_auto_archive_after, validate_default_thread_slowmode, validate_nsfw,
    validate_slowmode, validate_topic
)

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildTextBase(ChannelMetadataGuildMainBase):
    """
    Guild text channel metadata base.
    
    Attributes
    ----------
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    default_thread_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    default_thread_slowmode : `int`
        Applied as `thread.slowmode` when one is created.
    name : `str`
        The channel's name.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites :`None`,  `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    slowmode : `int`
        The amount of time in seconds that a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channels` permissions are unaffected.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('default_thread_auto_archive_after', 'default_thread_slowmode', 'nsfw', 'slowmode', 'topic')
    
    
    def __new__(
        cls,
        *,
        default_thread_auto_archive_after = ...,
        default_thread_slowmode = ...,
        name = ...,
        nsfw = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        slowmode = ...,
        topic = ...,
    ):
        """
        Creates a new guild text base channel metadata from the given parameters.
        
        Parameters
        ----------
        default_thread_auto_archive_after : `int``, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        default_thread_slowmode : `int`
            Applied as `thread.slowmode` when one is created.
        name : `str`, Optional (Keyword only)
            The channel's name.
        nsfw : `bool`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds that a user needs to wait between it's each message.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_thread_auto_archive_after
        if default_thread_auto_archive_after is ...:
            default_thread_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_thread_auto_archive_after = validate_default_thread_auto_archive_after(
                default_thread_auto_archive_after
            )
        
        # default_thread_slowmode
        if default_thread_slowmode is ...:
            default_thread_slowmode = SLOWMODE_DEFAULT
        else:
            default_thread_slowmode = validate_default_thread_slowmode(default_thread_slowmode)
        
        # nsfw
        if nsfw is ...:
            nsfw = False
        else:
            nsfw = validate_nsfw(nsfw)
        
        # slowmode
        if slowmode is ...:
            slowmode = SLOWMODE_DEFAULT
        else:
            slowmode = validate_slowmode(slowmode)
        
        # topic
        if topic is ...:
            topic = None
        else:
            topic = validate_topic(topic)
        
        # Construct
        self = ChannelMetadataGuildMainBase.__new__(
            cls,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        self.default_thread_auto_archive_after = default_thread_auto_archive_after
        self.default_thread_slowmode = default_thread_slowmode
        self.nsfw = nsfw
        self.slowmode = slowmode
        self.topic = topic
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            name = keyword_parameters.pop('name', ...),
            default_thread_auto_archive_after = keyword_parameters.pop('default_thread_auto_archive_after', ...),
            default_thread_slowmode = keyword_parameters.pop('default_thread_slowmode', ...),
            nsfw = keyword_parameters.pop('nsfw', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # default_thread_auto_archive_after
        hash_value ^= self.default_thread_auto_archive_after << 16
        
        # default_thread_slowmode
        hash_value ^= self.default_thread_slowmode << 8
        
        # nsfw
        hash_value ^= self.nsfw << 28
        
        # slowmode
        hash_value ^= self.slowmode << 4
        
        # topic
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # default_thread_auto_archive_after
        if self.default_thread_auto_archive_after != other.default_thread_auto_archive_after:
            return False
        
        # default_thread_slowmode
        if self.default_thread_slowmode != other.default_thread_slowmode:
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        # slowmode
        if self.slowmode != other.slowmode:
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildTextBase, cls)._create_empty()
        
        self.default_thread_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.default_thread_slowmode = SLOWMODE_DEFAULT
        self.nsfw = False
        self.slowmode = SLOWMODE_DEFAULT
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy)
    def copy(self):
        new = ChannelMetadataGuildMainBase.copy(self)
        new.default_thread_auto_archive_after = self.default_thread_auto_archive_after
        new.default_thread_slowmode = self.default_thread_slowmode
        new.nsfw = self.nsfw
        new.slowmode = self.slowmode
        new.topic = self.topic
        return new
    
    
    def copy_with(
        self,
        *,
        default_thread_auto_archive_after = ...,
        default_thread_slowmode = ...,
        name = ...,
        nsfw = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        slowmode = ...,
        topic = ...,
    ):
        """
        Copies the guild text base channel metadata with the given fields.
        
        Parameters
        ----------
        default_thread_auto_archive_after : `int``, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        default_thread_slowmode : `int`
            Applied as `thread.slowmode` when one is created.
        name : `str`, Optional (Keyword only)
            The channel's name.
        nsfw : `bool`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        slowmode : `int`, Optional (Keyword only)
            The amount of time in seconds that a user needs to wait between it's each message.
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # default_thread_auto_archive_after
        if default_thread_auto_archive_after is ...:
            default_thread_auto_archive_after = self.default_thread_auto_archive_after
        else:
            default_thread_auto_archive_after = validate_default_thread_auto_archive_after(
                default_thread_auto_archive_after
            )
        
        # default_thread_slowmode
        if default_thread_slowmode is ...:
            default_thread_slowmode = self.default_thread_slowmode
        else:
            default_thread_slowmode = validate_default_thread_slowmode(default_thread_slowmode)
        
        # nsfw
        if nsfw is ...:
            nsfw = self.nsfw
        else:
            nsfw = validate_nsfw(nsfw)
        
        # slowmode
        if slowmode is ...:
            slowmode = self.slowmode
        else:
            slowmode = validate_slowmode(slowmode)
        
        # topic
        if topic is ...:
            topic = self.topic
        else:
            topic = validate_topic(topic)
        
        # Construct
        new = ChannelMetadataGuildMainBase.copy_with(
            self,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        new.default_thread_auto_archive_after = default_thread_auto_archive_after
        new.default_thread_slowmode = default_thread_slowmode
        new.nsfw = nsfw
        new.slowmode = slowmode
        new.topic = topic
        return new
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            name = keyword_parameters.pop('name', ...),
            default_thread_auto_archive_after = keyword_parameters.pop('default_thread_auto_archive_after', ...),
            default_thread_slowmode = keyword_parameters.pop('default_thread_slowmode', ...),
            nsfw = keyword_parameters.pop('nsfw', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            slowmode = keyword_parameters.pop('slowmode', ...),
            topic = keyword_parameters.pop('topic', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # default_thread_auto_archive_after
        self.default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        
        # default_thread_slowmode
        self.default_thread_slowmode = parse_default_thread_slowmode(data)
        
        # nsfw
        self.nsfw = parse_nsfw(data)
        
        # slowmode
        self.slowmode = parse_slowmode(data)
        
        # topic
        self.topic = parse_topic(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # default_thread_auto_archive_after
        default_thread_auto_archive_after = parse_default_thread_auto_archive_after(data)
        if self.default_thread_auto_archive_after != default_thread_auto_archive_after:
            old_attributes['default_thread_auto_archive_after'] = self.default_thread_auto_archive_after
            self.default_thread_auto_archive_after = default_thread_auto_archive_after
        
        # default_thread_slowmode
        default_thread_slowmode = parse_default_thread_slowmode(data)
        if self.default_thread_slowmode != default_thread_slowmode:
            old_attributes['default_thread_slowmode'] = self.default_thread_slowmode
            self.default_thread_slowmode = default_thread_slowmode
        
        # nsfw
        nsfw = parse_nsfw(data)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        # slowmode
        slowmode = parse_slowmode(data)
        if self.slowmode != slowmode:
            old_attributes['slowmode'] = self.slowmode
            self.slowmode = slowmode
        
        # topic
        topic = parse_topic(data)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # default_auto_archive_duration
        put_default_thread_auto_archive_after_into(self.default_thread_auto_archive_after, data, defaults)
        
        # default_thread_slowmode
        put_default_thread_slowmode_into(self.default_thread_slowmode, data, defaults)
        
        # nsfw
        put_nsfw_into(self.nsfw, data, defaults)
        
        # slowmode
        put_slowmode_into(self.slowmode, data, defaults)
        
        # topic
        put_topic_into(self.topic, data, defaults)
        
        return data



__all__ = ('CommandBaseCustomId', )

from scarletio import copy_docs

from ..command_base import CommandBase


class CommandBaseCustomId(CommandBase):
    """
    Base type for commands based on `custom_id` matching.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``SlashCommand``
        The parent slasher of the component command.
    _parameter_converters : `tuple` of ``ParameterConverter``
        Parsers to parse command parameters.
    _command_function : `async-callable˛
        The command's function to call.
    _string_custom_ids : `None`, `tuple` of `str`
        The custom id-s to wait for.
    _regex_custom_ids : `None`, `tuple` of `re.Pattern`.
        Regex pattern to match custom-ids.
    name : `str`
        The command's name.
        
        Only used for debugging.
    
    response_modifier : `None`, ``ResponseModifier``
        Modifies values returned and yielded to command coroutine processor.
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    COMMAND_PARAMETER_NAMES : tuple of `str`
        All parameters names accepted by ``.__new__``
    COMMAND_NAME_NAME : `str`
        The command's "command" defining parameter's name.
    """
    __slots__ = (
        '_command_function', '_parameter_converters', '_regex_custom_ids', '_string_custom_ids', 'response_modifier',
    )
    
    COMMAND_PARAMETER_NAMES = (
        *CommandBase.COMMAND_PARAMETER_NAMES,
        'custom_id',
    )
    
    
    def __new__(cls, func, custom_id, name = None, **keyword_parameters):
        """
        Creates a new custom_id based command instance.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        func : `None`, `async-callable`
            The function used as the command when using the respective slash command.
        custom_id : `str`, (`list`, `set`) of `str`, `tuple` of (`str`, (`list`, `set`) of `str`)
            Custom id to match by the component command.
        name : `None`, `str` = `None`, Optional
            The name of the component command.
        **keyword_parameters : Keyword parameters
            Additional keyword parameters
        
        Other parameters
        ----------------
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, \
                `list` of (`str`, ``UserBase``, ``Role`` ), Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Returns
        -------
        self : ``CommandBaseCustomId``, ``Router``
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError
    
    
    @copy_docs(CommandBase._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in CommandBase._cursed_repr_builder(self):
            
            string_custom_ids = self._string_custom_ids
            if (string_custom_ids is not None):
                
                repr_parts.append(', string_custom_ids = [')
                index = 0
                limit = len(string_custom_ids)
                
                while True:
                    string_custom_id = string_custom_ids[index]
                    repr_parts.append(repr(string_custom_id))
                    
                    index += 1
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
                
                repr_parts.append(']')
            
            regex_custom_ids = self._regex_custom_ids
            if (regex_custom_ids is not None):
                
                repr_parts.append(', regex_custom_ids = [')
                index = 0
                limit = len(regex_custom_ids)
                
                while True:
                    regex_custom_id = regex_custom_ids[index]
                    repr_parts.append(repr(regex_custom_id.pattern.pattern))
                    
                    index += 1
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
                
                repr_parts.append(']')
            
            
            yield repr_parts
            
            response_modifier = self.response_modifier
            if (response_modifier is not None):
                repr_parts.append(', response_modifier = ')
                repr_parts.append(repr(response_modifier))
    
    
    async def invoke(self, client, interaction_event, regex_match):
        """
        Calls the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        regex_match : `None`, ``RegexMatch``
            The matched regex if applicable.
        """
        return
    
    
    @copy_docs(CommandBase.__hash__)
    def __hash__(self):
        hash_value = CommandBase.__hash__(self)
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except KeyError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _parameter_converters
        # Internal field
        
        # _regex_custom_ids
        regex_custom_ids = self._regex_custom_ids
        if (regex_custom_ids is not None):
            hash_value ^= hash(regex_custom_ids)
        
        # _string_custom_ids
        string_custom_ids = self._string_custom_ids
        if (string_custom_ids is not None):
            hash_value ^= hash(string_custom_ids)
        
        # response_modifier
        response_modifier = self.response_modifier
        if (response_modifier is not None):
            hash_value ^= hash(response_modifier)
        
        return hash_value
    
    
    @copy_docs(CommandBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBase._is_equal_same_type(self, other):
            return False
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _parameter_converters
        # Internal field
        
        # _regex_custom_ids
        if self._regex_custom_ids != other._regex_custom_ids:
            return False
        
        # _string_custom_ids
        if self._string_custom_ids != other._string_custom_ids:
            return False
        
        # response_modifier
        if self.response_modifier != other.response_modifier:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the command.
        
        Returns
        -------
        new : ``CommandBaseCustomId``
        """
        new = CommandBase.copy(self)
        
        # _command_function
        new._command_function = self._command_function
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # _regex_custom_ids
        new._regex_custom_ids = self._regex_custom_ids
        
        # _string_custom_ids
        new._string_custom_ids = self._string_custom_ids
        
        # response_modifier
        new.response_modifier = self.response_modifier
        
        return new
